import subprocess
import numpy as np
from .utils import split_dataset_into_stage, get_all_adj_adata
import os
import scanpy as sc
import gc
from .gcn_utilis import get_gcn_exp
from .UNAGI_runner import UNAGI_runner
import torch
from .pyro_models import VAE,Discriminator
from .UNAGI_analyst import analyst
from .trainer import UNAGI_trainer
class UNAGI():
    def __init__(self):
        super(UNAGI, self).__init__()
    def setup_data(self, data_path,stage_key,total_stage,gcn_connectivities=False,neighbors=25,threads = 20):
        self.data_folder = os.path.dirname(data_path)
        if os.path.exists(os.path.join(self.data_folder ,'0.h5ad')):
            temp = sc.read(os.path.join(self.data_folder , '0.h5ad'))
            self.input_dim = temp.shape[1]
            if 'gcn_connectivities' not in list(temp.obsp.keys()):
                gcn_connectivities = False
        else:
            print('The dataset is not splited into stages, please use setup_data function to split the dataset into stages first')
            self.data_path = data_path
            self.split_dataset_into_stage()
            gcn_connectivities = False
            
        self.data_path = os.path.join(data_path,'0.h5ad')
        self.stage_key = stage_key
        self.ns = total_stage
        #data folder is the folder that contains all the h5ad files
        self.data_folder = data_path#os.path.dirname(data_path)
        dir1 = os.path.join(self.data_folder , '0')
        dir2 = os.path.join(self.data_folder , '0/stagedata')
        dir3 = os.path.join(self.data_folder , 'model_save')
        initalcommand = 'mkdir '+ dir1 +' && mkdir '+dir2 +' && mkdir '+dir3
        p = subprocess.Popen(initalcommand, stdout=subprocess.PIPE, shell=True)

        if not gcn_connectivities:
            print('Cell graphs not found, calculating cell graphs for individual stages! Using K=%d and threads=%d for cell graph construction'%(neighbors,threads))
            self.calculate_neighbor_graph(neighbors,threads)
        
    def split_dataset_into_stage(self):
        split_dataset_into_stage(self.data_path, self.data_folder, self.stage_key)
    def calculate_neighbor_graph(self, neighbors=25,threads = 20):
        get_gcn_exp(self.data_folder, self.ns ,neighbors,threads= threads)
    def setup_training(self,
                 task, 
                 dist,
                 device=None,
                 epoch_iter=10,
                 epoch_initial=20,
                 lr=1e-4,
                 lr_dis = 5e-4,
                 beta=1,
                 hiddem_dim=256,
                 latent_dim=64,
                 graph_dim=1024,
                 BATCHSIZE=512,
                 max_iter=10,
                 GPU=False):
        self.dist = dist
        self.device = device
        self.epoch_iter = epoch_iter
        self.epoch_initial = epoch_initial
        self.lr = lr
        self.beta = beta
        self.lr_dis = lr_dis
        self.task = task
        self.latent_dim = latent_dim
        self.graph_dim = graph_dim
        self.hidden_dim = hiddem_dim
        self.BATCHSIZE = BATCHSIZE
        self.max_iter = max_iter
        self.GPU = GPU
        if self.GPU:
            assert self.device is not None, "GPU is enabled but device is not specified"
            self.device = torch.device(self.device)
        else:
            self.device = torch.device('cpu')

        self.model = VAE(self.input_dim, self.latent_dim, self.hidden_dim)#,beta=1,distribution=self.dist)
        # self.dis_model = Discriminator(self.input_dim)
        self.unagi_trainer = UNAGI_trainer(self.model,self.task,self.BATCHSIZE,self.epoch_initial,self.epoch_iter,self.device,self.lr, self.lr_dis,cuda=self.GPU)
    def run_UNAGI(self,idrem_dir):
        for iteration in range(0,self.max_iter):
            
            if iteration != 0:
                dir1 = os.path.join(self.data_folder , str(iteration))
                dir2 = os.path.join(self.data_folder , str(iteration)+'/stagedata')
                dir3 = os.path.join(self.data_folder , 'model_save')
                initalcommand = 'mkdir '+ dir1 +' && mkdir '+dir2
                p = subprocess.Popen(initalcommand, stdout=subprocess.PIPE, shell=True)
            unagi_runner = UNAGI_runner(self.data_folder,self.ns,iteration,self.unagi_trainer,idrem_dir)
            unagi_runner.run()

  
    def analyse_UNAGI(self,data_path,iteration,progressionmarker_background_sampling_times,target_dir=None,customized_drug=None,cmap_dir=None):
        analysts = analyst(data_path,iteration,target_dir=target_dir,customized_drug=customized_drug,cmap_dir=cmap_dir)
        analysts.start_analyse(progressionmarker_background_sampling_times)
        print('analysed finished please check the outputs!')


        

        


        




from UNAGI import UNAGI
unagi = UNAGI()
unagi.setup_data('/mnt/md0/yumin/to_upload/UNAGI/data/small/0.h5ad',total_stage=4,stage_key='stage',splited_dataset=True)
unagi.setup_training(task='small',dist='ziln',device='cuda:0',GPU=True,epoch_iter=15,epoch_initial=10,max_iter=3,BATCHSIZE=560)
unagi.run_UNAGI()


#this is push back, last cell is push forward
import scipy
import time 
import scanpy as sc
import numpy as np
import os
import random
import gc
import torch
from scipy.sparse import issparse
from scipy.sparse import csr_matrix
#import DataLoader from torch
from torch.utils.data import DataLoader
import threading
from .pyro_models import VAE
from .analysis_perturbation import perturbationAnalysis
class perturbation:
    def __init__(self, target_directory,model_name,idrem_dir):
        self.model_name = model_name
        self.target_directory = target_directory
        
        self.idrem_dir = idrem_dir
        

        self.adata = self.read_mergeadata()
        self.total_stage = len(set(self.adata.obs['stage']))
        self.tracks = self.getTrackReadOrder()
        self.stageadata = self.read_stagedata()
        
        
        self.hiddenReps = []
        self.perturb_stage_data_mean = []

        
    def read_mergeadata(self):
        read_path = self.target_directory#os.path.join(self.target_directory,'dataset.h5ad')
        mergeadata = read_path#sc.read_h5ad(read_path)
        mergeadata.obs['leiden'] = mergeadata.obs['leiden'].astype('string')
        return mergeadata
    def read_stagedata(self):
        stageadata = []
        self.stage_cluster = {}
        stage_have_clusters = {}
        stage_have_clusters[str(0)] = []
        for i in self.tracks.keys():
            track = self.getTrack(len(self.adata.obs['stage'].unique())-1,i)
            track_name = str(track[0][0])
            stage_have_clusters[str(0)].append(str(track[0][0]))
            for j in range(1,len(track)):
                if str(j) not in stage_have_clusters.keys():
                    stage_have_clusters[str(j)] = []
                stage_have_clusters[str(j)].append(str(track[j][0]))
                track_name += '-' + str(track[j][0])
        self.adata.obs['stage'] = self.adata.obs['stage'].astype('string')
        for i in range(0,4):

            stagedataids = self.adata.obs[self.adata.obs['stage']==str(i)].index.values
            

            adata = self.adata[stagedataids]
            adata.obs['leiden'] = adata.obs['leiden'].astype('string')
            grouped = adata.obs.groupby('leiden')
            self.stage_cluster[str(i)] = {}#{name: adata[group.index.tolist()] for name, group in grouped}
            for name, group in grouped:
                if str(name) not in stage_have_clusters[str(i)]:
                    continue
                else:
                    self.stage_cluster[str(i)][str(name)] = adata[group.index.tolist()]
            stageadata.append(adata)
        return stageadata
    def get_gene_names(self):
        return np.array(list(self.stageadata[0].var.index.values))
    # def save_to_disk(self):
    #         for i, each in enumerate(self.stageadata):
    #             save_path = os.path.join(self.target_directory,'stagedata/concat_%d.h5ad'%(i))
    #             each.write_h5ad(save_path)
    def getDescendants(tempcluster,stage,edges):
        out = []
        for each in tempcluster:
            
            for item in edges[str(stage-1)]:
            
                if each == item[0]:
                    
                    out.append(item[1])
        return out
    def getDistance(self,rep, cluster):

        cluster = cluster.reshape(1,-1)
        cluster = cluster.repeat(rep.shape[0],axis=0)
        return np.linalg.norm(rep-cluster,axis=1)
    def matchSingleClusterGeneDict(self,goadata,gotop):
        gene_dict={}
        for i,each in enumerate(goadata.var.index.tolist()):
            gene_dict[each]=i
        results=[]
        for each in gotop:
            if each not in goadata.var.index.tolist():

                continue
            results.append(gene_dict[each])
        return results
    def getTrack(self,stage, clusterid):
        path = self.idrem_dir#os.path.join(self.target_directory,'idremVizCluster')
        filenames = os.listdir(path) #defalut path

        tempTrack = [[] for _ in range(self.total_stage)]
        for each in filenames:
            temp = each.split('.')[0].split('-')
            for i,item in enumerate(temp):
                temp1 = item.split('n')
                tempTrack[i].append(temp1)
        track = [[] for _ in range(self.total_stage)]
        # open_path = os.path.join(self.target_directory,'edges.txt')
        # edges = eval(open(open_path, 'r').read())
        edges = self.adata.uns['edges']
        for i, each in enumerate(tempTrack[int(stage)]):
            if str(clusterid) in each:
                track[0] = [int(tempTrack[0][i][0])]
                
                tempcluster = clusterid
                for k in range(int(stage),0, -1):
                    for new_each in edges[str(k-1)]:
                        if new_each[1] == tempcluster:
                            track[k]= [new_each[1]]#[edges[str(k-1)][tempcluster][1]]
                            tempcluster = new_each[0] #edges[str(k-1)][tempcluster][0] 
                            break
                    # track[k]= [edges[str(k-1)][tempcluster][1]]
                    
                tempcluster = [clusterid]
                
                for k in range(int(stage)+1,self.total_stage):
                    
                    track[k]=self.getDescendants(tempcluster,k,edges)
                    tempcluster=track[k]
        return track

    def getZandZc(self,adata,stage,cluster,CUDA = False, impactfactor=None,topN = None):
        if stage < len(self.hiddenReps):
            if impactfactor is None:
                return self.hiddenReps[stage]
            else:
                data = self.perturb_stage_data_mean[stage]
        else:
            
            

            clusterAdataID = adata.obs[adata.obs['leiden'] == str(cluster)].index.tolist()
            clusterAdata = adata[clusterAdataID]
            
            # gcn_connectivities = clusterAdata.obsp['gcn_connectivities']
            
            # temp = gcn_connectivities[clusterAdataID]
            # temp = list(set(temp.indices))
            # # input = np.zeros(shape = (adata.X.shape[0],adata.X.shape[1]))
            # input = csr_matrix(input)
                
            # input[temp] = adata.X[temp]
            # input = gcn_connectivities @ input
            # input = input[temp].toarray()
            input = clusterAdata.obsp['gcn_connectivities'] @ clusterAdata.X
            if issparse(input):
                input = input.toarray()
            data = np.mean(input,axis=0).reshape(1,-1)
        loadModelDict = self.model_name#'./'+self.target_directory+'/model_save/'+self.model_name+'.pth'
        # vae = VAE(clusterAdata.shape[1], 256, 1024, 64,beta=1,distribution='ziln')#torch.load(loadModelDict)
        # vae = torch.load(loadModelDict)
        vae = VAE(data.shape[1], 64, 256, 0.5)
        if CUDA:
            vae.load_state_dict(torch.load(loadModelDict,map_location=torch.device('cuda:0')))
            # vae = torch.load(loadModelDict)
            vae.to('cuda:0')
        else:

            vae.load_state_dict(torch.load(loadModelDict), map_location=torch.device('cpu'))
            vae.to('cpu')
        
        vae.eval()
        recons = []
        zs = []
        zmeans = []
        zstds = []
        # data = np.mean(clusterAdata.layers['concat'].toarray(),axis=0).reshape(1,-1)#
        # print(np.mean(data))
        if impactfactor is not None:
            data = np.repeat(data.reshape(1,-1), len(impactfactor), axis=0)
            cell_loader = DataLoader(data.astype('float32'), batch_size=len(impactfactor), shuffle=False, num_workers=0)
        else:
            cell_loader = DataLoader(data.astype('float32'), batch_size=1, shuffle=False, num_workers=0)
        # cell_loader = DataLoader(clusterAdata.layers['concat'].toarray().astype('float32'), batch_size=2000, shuffle=False, num_workers=1)
        # xx_adj = np.ones(np.shape(clusterAdata.X)[0])
        
        adj = torch.tensor(1)
        if CUDA:
            adj = adj.to('cuda:0')
        for x in cell_loader:


            if impactfactor is not None:

                impactfactor = impactfactor.astype('float32')
                # x = x+impactfactor
                x = x+x*impactfactor

    
            # if on GPU put mini-batch into CUDA memory
            if CUDA:
                x = x.to('cuda:0')

            z=vae.getZ(x,adj,0,0,1,test=True)
            zs+=z[0].cpu().detach().numpy().tolist()
            zmeans+=z[1].cpu().detach().numpy().tolist()
            zstds+=z[2].cpu().detach().numpy().tolist()
        zs = np.array(zs)

        if stage >= len(self.hiddenReps):
            self.hiddenReps.append(zs)
            self.perturb_stage_data_mean.append(data)
        return zs
    class perturbationthread(threading.Thread):
        def __init__(self, outer_instance,outs, selectedstage,selectedcluster,track,bound,perturbated_gene,CUDA):
            threading.Thread.__init__(self)
            self.selectedstage = selectedstage
            self.selectedcluster = selectedcluster
            self.track = track
            self.bound = bound
            self.outs = outs
            self.perturbated_gene = perturbated_gene
            self.outer_instance = outer_instance
            self.CUDA = CUDA
        def run(self):
            self.outs[self.selectedstage]+=self.outer_instance.perturbation__auto_centroid(self.outer_instance.stageadata[self.selectedstage], self.outer_instance.stageadata, self.selectedstage, self.selectedcluster, self.track, self.bound,self.perturbated_gene,self.CUDA)

    def getTrackReadOrder(self):
        '''
        for each completed path in track (completed path = control->1->2->3, number of completed paths = number of 3 nodes), return a dictionary of orders. 
        like the path has stage3:1 is the second one to be read.
        '''
        path = self.idrem_dir#os.path.join(self.target_directory,'idremVizCluster')
        filenames = os.listdir(path) #defalut path
        tempTrack = [[] for _ in range(self.total_stage)]
        for each in filenames:
            temp = each.split('.')[0].split('-')
            for i,item in enumerate(temp):
                temp1 = item.split('n')
                tempTrack[i].append(temp1)
        dic = {}
        for i, ids in enumerate(tempTrack[-1]):
            for each in ids:
                dic[int(each)] = i
        return dic
    def perfect_perturbation__auto_centroid(self,stageadata, adata, selectedstage,selectedcluster,track,bound,perturbated_genes,CUDA=False):
        '''
        remove non top genes and tf. compared
        '''

        hiddenReps=[]
        repNodes = []
        flag = -1 #to indicate which cluster in the track to be changed like [40,1,3,2,4,5] 3 is no.2
        clusterids = []
        zc = []
        adatacollection = adata#[]
        
        plotadata = []
        for stage, clusters in enumerate(track):
            for clusterid, leiden in enumerate(clusters):
                if stage == selectedstage and leiden == selectedcluster:
                    flag = len(hiddenReps)
                
                temp = self.getZandZc(adatacollection[stage],stage,leiden,CUDA=CUDA)
            
                hiddenReps.append(temp)
        hiddenReps = np.array(hiddenReps)

        dijresults = []
        
        count=0
        
        for stage, clusters in enumerate(track):
            temp = []
            for clusterid, leiden in enumerate(clusters):
            
                temp=self.getDistance(hiddenReps[flag],hiddenReps[count])
                count+=1
            dijresults.append(temp)
            

        perturbated_stage = int(selectedstage)
  
        
        
        impactFactor = perturbated_genes
        impactFactor[impactFactor!=0] = 1
        temp = impactFactor[0].copy()
        perturbated_stage = int(selectedstage)
        selectedstage=1
        if bound > 0:
            mean_previous = np.mean(np.array(adatacollection[selectedstage-1].X.toarray()),axis=0)
            mean_current = np.mean(np.array(adatacollection[selectedstage].X.toarray()),axis=0)
            if selectedstage == 0:
                mean_previous = mean_current
            
            diff = mean_previous - mean_current
            por = diff/mean_current
            impactFactor = por * impactFactor
        else:

            mean_previous = np.mean(np.array(adatacollection[selectedstage+1].X.toarray()),axis=0)
            mean_current = np.mean(np.array(adatacollection[selectedstage].X.toarray()),axis=0)
            if selectedstage == len(adata)-1:
                mean_previous = mean_current
            diff = mean_previous - mean_current
            por = diff/mean_current
            impactFactor = por * impactFactor

        selectedtemp = self.getZandZc(None,selectedstage,track[selectedstage][0],impactfactor = impactFactor,CUDA=CUDA)
        count = 0
        fijresults = []
        for stage, clusters in enumerate(track):
            temp = []
            for clusterid, leiden in enumerate(clusters):
                temp = self.getDistance(selectedtemp,hiddenReps[count])
                count+=1
            fijresults.append(temp)
        delta = np.array(fijresults) - np.array(dijresults)

        gc.collect()
        
        out = []
        for i in range(impactFactor.shape[0]):
            temp = []
            temp.append(perturbated_genes[i])
            for kk in range(len(track)):
                temp.append(track[kk][0])
            for kk in range(len(track)):
                temp.append(delta[kk][i])
            out.append(temp)

        return out
    def perturbation__auto_centroid(self,stageadata, adata, selectedstage,selectedcluster,track,bound,perturbated_genes,CUDA=False):
        '''
        remove non top genes and tf. compared
        '''

        hiddenReps=[]
        repNodes = []
        flag = -1 #to indicate which cluster in the track to be changed like [40,1,3,2,4,5] 3 is no.2
        clusterids = []
        zc = []
        adatacollection = adata#[]
        
        plotadata = []
        for stage, clusters in enumerate(track):
            for clusterid, leiden in enumerate(clusters):
                if stage == selectedstage and leiden == selectedcluster:
                    flag = len(hiddenReps)
               
                temp = self.getZandZc(adatacollection[stage],stage,leiden,CUDA=CUDA)
            
                hiddenReps.append(temp)
        hiddenReps = np.array(hiddenReps)

        dijresults = []
        
        count=0
        
        for stage, clusters in enumerate(track):
            temp = []
            for clusterid, leiden in enumerate(clusters):
            
                temp=self.getDistance(hiddenReps[flag],hiddenReps[count])
                count+=1
            dijresults.append(temp)
        impactFactor = perturbated_genes
        selectedtemp = self.getZandZc(None,selectedstage,track[selectedstage][0],impactfactor = impactFactor,CUDA=CUDA)
        count = 0
        fijresults = []
        for stage, clusters in enumerate(track):
            temp = []
            for clusterid, leiden in enumerate(clusters):
                temp = self.getDistance(selectedtemp,hiddenReps[count])
                count+=1
            fijresults.append(temp)
        delta = np.array(fijresults) - np.array(dijresults)

        gc.collect()
        
        out = []
        for i in range(impactFactor.shape[0]):
            temp = []
            temp.append(perturbated_genes[i])
            for kk in range(len(track)):
                temp.append(track[kk][0])
            for kk in range(len(track)):
                temp.append(delta[kk][i])
            out.append(temp)

        return out
    def prepare_speed_perturbation_data(self, adata, stage, leiden,raw=None, impactfactors=None):
        if raw is None:
            stageadata = adata[stage]
            clusterAdata = self.stage_cluster[str(stage)][str(leiden)]#stageadata[self.stage_cluster_index[str(stage)][str(leiden)]]#.index.tolist()
            # clusterAdata = stageadata[clusterAdataID]
            # print(clusterAdata.obsp['gcn_connectivities'])
            # matrix1 = cupyx.scipy.sparse.csr_matrix(clusterAdata.obsp['gcn_connectivities'])#cp.asarray(clusterAdata.obsp['gcn_connectivities'])
            # matrix2 = cp.asarray(clusterAdata.X)
            # input = matrix1 @ matrix2
            # input = cp.asnumpy(input)
            input = clusterAdata.obsp['gcn_connectivities'] @ clusterAdata.X
            if issparse(input):
                input = input.toarray()

            data = np.mean(input,axis=0).reshape(1,-1)
        else:
            data = raw
        if impactfactors is not None:
            # print(np.mean(data + data * impactfactors- data))
            data =data + data * impactfactors
        
        if raw is None:
            return data.reshape(-1)
        else:
            return data
    def getZ_speedup(self, input,CUDA=False):
        zs = []
        zmeans = []
        zstds = []
        loadModelDict = self.model_name#'./'+self.target_directory+'/model_save/'+self.model_name+'.pth'
        vae = myGrphVAEPyroModule(input.shape[1], 64, 0, 0.5)
        if CUDA:
            
            vae.load_state_dict(torch.load(loadModelDict,map_location='cuda:0'))
            # vae = torch.load(loadModelDict)
            vae.to('cuda:0')
        else:
            vae.load_state_dict(torch.load(loadModelDict), map_location=torch.device('cpu'))
            vae.to('cpu')
        
        vae.eval()
        cell_loader = DataLoader(input.astype('float32'), batch_size=len(input), shuffle=False, num_workers=0)
        adj = torch.tensor(1)
        if CUDA:
            adj = adj.to('cuda:0')
        for x in cell_loader:
            # if on GPU put mini-batch into CUDA memory
            if CUDA:
                x = x.to('cuda:0')
            z=vae.getZ(x,adj,0,0,1,test=True)
            zs+=z[0].cpu().detach().numpy().tolist()
            zmeans+=z[1].cpu().detach().numpy().tolist()
            zstds+=z[2].cpu().detach().numpy().tolist()
        zs = np.array(zs)
        return zs
    def perturbation__auto_centroid_speed(self,adata, lastClusters, perturbated_genes,CUDA=False):
        '''
        remove non top genes and tf. compared
        '''

        hiddenReps=[]
        repNodes = []
        flag = -1 #to indicate which cluster in the track to be changed like [40,1,3,2,4,5] 3 is no.2
        clusterids = []
        zc = []
        input_data = []
        input_pertubred = []
        # adatacollection = adata#[]
        
        for i, each in enumerate(lastClusters):
            track = self.getTrack(len(self.stageadata)-1,each)
            for stage, clusters in enumerate(track):
                for clusterid, leiden in enumerate(clusters):
                    input_data.append(self.prepare_speed_perturbation_data(adata, stage, leiden))
        
        input_data = np.array(input_data)
        input_pertubred_forward = self.prepare_speed_perturbation_data(adata, stage, leiden, raw =input_data, impactfactors = perturbated_genes[0])
        input_pertubred_backward = self.prepare_speed_perturbation_data(adata, stage, leiden, raw =input_data, impactfactors = perturbated_genes[1])
        input_pertubred = np.append(input_pertubred_forward,input_pertubred_backward,axis=0)
        input_pertubred = np.array(input_pertubred)

        Z_input = self.getZ_speedup(input_data,CUDA).reshape(-1,4,64)
        Z_perturbed = self.getZ_speedup(input_pertubred,CUDA).reshape(-1,4,64)
        input_distance = []
        for i, each in enumerate(Z_input):
            for  j, each1 in enumerate(each):
                for k, each2 in enumerate(each):
                    each1 = each1.reshape(1, -1)
                    each2 = each2.reshape(1, -1)
                    input_distance.append(self.getDistance(each1,each2))
        input_distance = np.array(input_distance)
        input_distance = input_distance.reshape(-1,4,4)
        distance = []
        # print(input_distance.shape)
        for i, each in enumerate(Z_perturbed):
            for j, each1 in enumerate(Z_perturbed[i]):
                if i < len(Z_perturbed)//2:
                    count = i
                else:
                    count = i - len(Z_perturbed)//2
                for k, each2 in enumerate(Z_input[count]):
                    each1 = each1.reshape(1, -1)
                    each2 = each2.reshape(1, -1)
                    distance.append(self.getDistance(each1,each2))
        
        distance = np.array(distance)
        distance = distance.reshape(-1,4,4)
        # print(distance.shape)
        
        #delta1 is first half of delta, delta2 is second half of delta
        delta1 = distance[:len(distance)//2]- input_distance
        delta2 = distance[len(distance)//2:]- input_distance
        # print(delta1.shape)
        # print(delta2.shape)
        return delta1, delta2
        
        # out = []
        # for i in range(impactFactor.shape[0]):
        #     temp = []
        #     temp.append(perturbated_genes[i])
        #     for kk in range(len(track)):
        #         temp.append(track[kk][0])
        #     for kk in range(len(track)):
        #         temp.append(delta[kk][i])
        #     out.append(temp)

        return out
    def get_drug_genes(self,bound):
        drug_target = self.adata.uns['data_drug_overlap_genes']#this should be an attribute of adata object later on
        #drug_target = dict(np.load(drug_cell_type_target,allow_pickle=True).tolist()) #this should be an attribute of adata object later on
        drug_names = list(drug_target.keys())

        drug_target_genes = list(drug_target.values())
        temp_drug_target_genes = drug_target_genes.copy()
        perturbed_genes = []
        
        for temp in temp_drug_target_genes:
            
            out_temp = []
            for each in temp:
                each = each.split(':')
                if each[1] == '+':
                    each = each[0]+':'+str(bound)
                elif each[1] == '-':
                    each = each[0]+':'+str(1/bound)
                out_temp.append(each)
            perturbed_genes.append(out_temp)
        return drug_names, perturbed_genes
    def startAutoPerturbation_online(self,lastCluster,perturbed_genes,CUDA=False):
        out = {}
        out['online'] = {}
        track = self.getTrack(len(self.stageadata)-1,lastCluster)

        perturbed_genes = [perturbed_genes]
        perturbated_gene_ids = []
        impactFactor = []
        for perturbated_gene in perturbed_genes:
            
            temp_perturbated_gene = perturbated_gene.copy()
            temp_bound = []
            perturbated_gene = []
            for each in temp_perturbated_gene:
                each = each.split(':')
                if len(each) > 1:
                    temp_bound.append(float(each[1]))
                    perturbated_gene.append(each[0])
                else:
                    perturbated_gene = temp_perturbated_gene
                    break

            perturbated_gene_id = self.matchSingleClusterGeneDict(self.stageadata[-1],perturbated_gene)
            perturbated_gene_ids.append(perturbated_gene_id)
            temp = np.zeros(shape=(len(self.stageadata[-1].var.index.tolist())))

            for id_each, each in enumerate(perturbated_gene_id):
                temp[each] = temp_bound[id_each]-1
            
            impactFactor.append(temp)
        impactFactor = np.array(impactFactor)

  

        perturbation_results = [[] for i in range(len(track))]
   
        for i, selectedcluster in enumerate(track):
            # print(i)
            threads = []
            self.stageadata[i].obs['leiden'] = self.stageadata[i].obs['leiden'].astype('string')
            perturbation_results[i] += self.perturbation__auto_centroid(self.stageadata[i], self.stageadata, i, selectedcluster[0], track, None,impactFactor,CUDA)
                                      
            for thread in threads:
                thread.start()
            for thread in threads:
                thread.join()
            
            for od, each in enumerate(perturbation_results[i]):
                tempout = []
                for kk in range(self.total_stage):
                    tempout.append(each[len(each)-self.total_stage+kk])
                out['online'][str(i)] = tempout

        return out

    def startAutoPerturbation_online_speed(self,lastClusters,perturbed_genes,track_names,CUDA=False):
        out = {}
        out['online'] = {}
        
        temp_track = self.getTrack(len(self.stageadata)-1,lastClusters[0])
        impactFactors = []
        for each_perturbed_genes in perturbed_genes:
            perturbed_genes_temp = [each_perturbed_genes]
            perturbated_gene_ids = []
            impactFactor = []
            for perturbated_gene in perturbed_genes_temp:
                
                temp_perturbated_gene = perturbated_gene.copy()
                temp_bound = []
                perturbated_gene = []
                for each in temp_perturbated_gene:
                    each = each.split(':')
                    if len(each) > 1:
                        temp_bound.append(float(each[1]))
                        perturbated_gene.append(each[0])
                    else:
                        perturbated_gene = temp_perturbated_gene
                        break

                perturbated_gene_id = self.matchSingleClusterGeneDict(self.stageadata[-1],perturbated_gene)
                perturbated_gene_ids.append(perturbated_gene_id)
                temp = np.zeros(shape=(len(self.stageadata[-1].var.index.tolist())))

                for id_each, each in enumerate(perturbated_gene_id):
                    temp[each] = temp_bound[id_each]-1
                
                impactFactor.append(temp)
            impactFactor = np.array(impactFactor)
            impactFactors.append(impactFactor)
  

        perturbation_results = [[] for i in range(len(temp_track))]
        outs = self.perturbation__auto_centroid_speed(self.stageadata,lastClusters,impactFactors,CUDA)
        out1 = {}
        out2 = {}
        for i, each in enumerate(lastClusters):
            track = self.getTrack(len(self.stageadata)-1,each)

            out1[track_names[i]] = {}
            out2[track_names[i]] = {}
            out1[track_names[i]]['online'] = {}
            out2[track_names[i]]['online'] = {}
            for j, selectedcluster in enumerate(track):
                out1[track_names[i]]['online'][str(j)] = outs[0][i][j]
                out2[track_names[i]]['online'][str(j)] = outs[1][i][j]
        
        return out1, out2

        # for i, selectedcluster in enumerate(track):
        #     # print(i)
        #     threads = []
        #     self.stageadata[i].obs['leiden'] = self.stageadata[i].obs['leiden'].astype('string')
        #     perturbation_results[i] += self.perturbation__auto_centroid(self.stageadata[i], self.stageadata, i, selectedcluster[0], track, None,impactFactor,CUDA)
                                      
        #     for thread in threads:
        #         thread.start()
        #     for thread in threads:
        #         thread.join()
            
        #     for od, each in enumerate(perturbation_results[i]):
        #         tempout = []
        #         for kk in range(self.total_stage):
        #             tempout.append(each[len(each)-self.total_stage+kk])
        #         out['online'][str(i)] = tempout

        # return out
    
    def assign_random_direction_to_random_genes(self,random_genes):
        out = []
        reversed_out = []
        temp_random_genes = random_genes.copy()
        for temp in temp_random_genes:
            out_temp = []
            copyout_temp = []
            for each in temp:
                copyeach = each
                flag = np.random.choice(['-','+'])
                # bound = np.random.uniform(1.00001,3)
                bound = 3
                if flag == '+':
                    
                    each = each+':'+str(bound)
                    copyeach = copyeach+':'+str(1/bound)
                elif flag == '-':
                    each = each+':'+str(1/bound)
                    copyeach = copyeach+':'+str(bound)
                out_temp.append(each)
                copyout_temp.append(copyeach)
            out.append(out_temp)
            reversed_out.append(copyout_temp)
        return out,reversed_out
   #~~~~~         
    def startAutoPerturbation(self,lastCluster,bound,mode,CUDA = True,random_genes= None, random_times = None,written=True):
        track = self.getTrack(len(self.stageadata)-1,lastCluster)

        track_name = str(track[0][0]) 
        for i in range(1,len(track)):
            track_name += '-' + str(track[i][0])
        outs = [[] for i in range(len(track))]
        if mode == 'drug':
            perturbed_items, perturbed_genes = self.get_drug_genes(bound)
        elif mode == 'pathway':
            pathway_gene = self.adata.uns['data_pathway_overlap_genes']
            #pathway_gene = dict(np.load('./'+self.target_directory+'/data_pathway_overlap_genes.npy',allow_pickle=True).tolist())#this should be an attribute of adata object later on
            perturbed_items = list(pathway_gene.keys())
            temp_perturbed_genes = list(pathway_gene.values())
            perturbed_genes = []
            for each in temp_perturbed_genes:
                if type(each)!= list:
                    each = each.tolist()

                perturbed_genes.append(each)
        
        ######
        elif mode == 'perfect':
            perturbed_items, perturbed_genes = self.get_drug_genes(bound)

        elif mode == 'random_background':
        
            perturbated_genes = []
            genelen = len(self.stageadata[0].var.index.tolist())
            genenames = np.array(list(self.stageadata[0].var.index.values))
            shuffled_gene_id = [j for j in range(genelen)]
        elif mode == 'online_random_background':
            bound = 'A'
            genelen = len(self.stageadata[0].var.index.tolist())
            genenames = np.array(list(self.stageadata[0].var.index.values))
            shuffled_gene_id = [j for j in range(genelen)]
            perturbed_genes = []
            perturbed_items = []
            
            for each in range(random_times):
                random_genes = random.randint(1,3)
                perturbed_items.append(str(each))
                random.shuffle(shuffled_gene_id)
                shuffled_gene = genenames[shuffled_gene_id[:random_genes]]
                perturbed_genes.append(shuffled_gene.tolist())
            perturbed_genes,reversed_perturbed_genes = self.assign_random_direction_to_random_genes(perturbed_genes)
        if mode != 'random_background':
            impactFactor = []
            perturbated_gene_ids = []
            for perturbated_gene in perturbed_genes:

                if type(perturbated_gene) != list:
                    perturbated_gene = [perturbated_gene]
                temp_perturbated_gene = perturbated_gene.copy()
                temp_bound = []
                perturbated_gene = []
                for each in temp_perturbated_gene:
                    each = each.split(':')
                    if len(each) > 1:
                        temp_bound.append(float(each[1]))
                        perturbated_gene.append(each[0])
                    else:
                        perturbated_gene = temp_perturbated_gene
                        break

                perturbated_gene_id = self.matchSingleClusterGeneDict(self.stageadata[-1],perturbated_gene)
                perturbated_gene_ids.append(perturbated_gene_id)
                temp = np.zeros(shape=(len(self.stageadata[-1].var.index.tolist())))
                if len(temp_bound) == 0:
                    temp[perturbated_gene_id] = 1
                    temp = temp*(bound-1)
                else:
                    for id_each, each in enumerate(perturbated_gene_id):
                        temp[each] = temp_bound[id_each]-1

                impactFactor.append(temp)
            impactFactor = np.array(impactFactor)


        for i, selectedcluster in enumerate(track):
    
            if '%s_perturbation_deltaD'%mode not in self.adata.uns.keys():
                self.adata.uns['%s_perturbation_deltaD'%mode] = {}
            threads = []
            if mode == 'random_background':
                perturbed_genes = []
                perturbed_items = []
                for each in range(random_times):
                    perturbed_items.append(str(each))
                    random.shuffle(shuffled_gene_id)
                    shuffled_gene = genenames[shuffled_gene_id[:random_genes]]
                    perturbed_genes.append(shuffled_gene.tolist())
                impactFactor = []
                perturbated_gene_ids = []
                for perturbated_gene in perturbed_genes:
                
                    if type(perturbated_gene) != list:
                        perturbated_gene = [perturbated_gene]
                    temp_perturbated_gene = perturbated_gene.copy()
                    temp_bound = []
                    perturbated_gene = []
                    for each in temp_perturbated_gene:
                        each = each.split(':')
                        if len(each) > 1:
                            temp_bound.append(float(each[1]))
                            perturbated_gene.append(each[0])
                        else:
                            perturbated_gene = temp_perturbated_gene
                            break

                    perturbated_gene_id = self.matchSingleClusterGeneDict(self.stageadata[-1],perturbated_gene)
                    perturbated_gene_ids.append(perturbated_gene_id)
                    temp = np.zeros(shape=(len(self.stageadata[-1].var.index.tolist())))
                    if len(temp_bound) == 0:
                        temp[perturbated_gene_id] = 1
                        temp = temp*(bound-1)
                    else:
                        for id_each, each in enumerate(perturbated_gene_id):
                            temp[each] = temp_bound[id_each]-1
                    
                    impactFactor.append(temp)
                impactFactor = np.array(impactFactor)

            self.stageadata[i].obs['leiden'] = self.stageadata[i].obs['leiden'].astype('string')
            # outs[i] = self.perturbationthread(self,i, selectedcluster[0], track, bound,impactFactor,CUDA)
            if mode == 'perfect':
                outs[i] +=self.perfect_perturbation__auto_centroid(self.stageadata[i], self.stageadata, i, selectedcluster[0], track, bound,impactFactor,CUDA)

            else:
                outs[i] += self.perturbation__auto_centroid(self.stageadata[i], self.stageadata, i, selectedcluster[0], track, bound,impactFactor,CUDA)

            for od, each in enumerate(outs[i]):
                if written == True:
                    # with open('./'+self.target_directory+'/tsdg2_%s_%s.csv'%(mode,bound),"a+") as f:
                        # f.write('%s,%d-%d-%d-%d,%.7f,%.7f,%.7f,%.7f\n'%(perturbed_items[od],each[1], each[2], each[3], each[4],each[5],each[6],each[7],each[8]))
                    pass #fix this later on
                else:
                    
                    # for  name in perturbed_items:
                    if str(bound) not in self.adata.uns['%s_perturbation_deltaD'%mode].keys():
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)] = {}
                    if track_name not in self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)].keys():
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name] = {}
                    if perturbed_items[od] not in self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name].keys():
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name][perturbed_items[od]] = {}
                    if str(i) not in self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name][perturbed_items[od]].keys():
                        tempout = []
                        for kk in range(self.total_stage):
                            tempout.append(each[len(each)-self.total_stage+kk])
                        # if self.mode == 'random_background':
                        #     if bound < 1/bound:
                        #         bound = 1/bound
                        #     self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name][perturbed_items[od]+str(bound)][str(i)] = tempout
                        # else:
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name][perturbed_items[od]][str(i)] = tempout
        if mode == 'online_random_background':
            bound = 'B'
            self.hiddenReps = []
            self.perturb_stage_data_mean = []
            reversed_perturbed_genes
            impactFactor = []
            perturbated_gene_ids = []
            for perturbated_gene in reversed_perturbed_genes:
            
                if type(perturbated_gene) != list:
                    perturbated_gene = [perturbated_gene]
                temp_perturbated_gene = perturbated_gene.copy()
                temp_bound = []
                perturbated_gene = []
                for each in temp_perturbated_gene:
                    each = each.split(':')
                    if len(each) > 1:
                        temp_bound.append(float(each[1]))
                        perturbated_gene.append(each[0])
                    else:
                        perturbated_gene = temp_perturbated_gene
                        break

                perturbated_gene_id = self.matchSingleClusterGeneDict(self.stageadata[-1],perturbated_gene)
                perturbated_gene_ids.append(perturbated_gene_id)
                temp = np.zeros(shape=(len(self.stageadata[-1].var.index.tolist())))
                if len(temp_bound) == 0:
                    temp[perturbated_gene_id] = 1
                    temp = temp*(bound-1)
                else:
                    for id_each, each in enumerate(perturbated_gene_id):
                        temp[each] = temp_bound[id_each]-1
                
                impactFactor.append(temp)
            impactFactor = np.array(impactFactor)
        
            outs = [[] for i in range(len(track))]
            for i, selectedcluster in enumerate(track):
                threads = []
                outs[i] += self.perturbation__auto_centroid(self.stageadata[i], self.stageadata, i, selectedcluster[0], track, bound,impactFactor,CUDA)
                # threads.append(self.perturbationthread(self,outs, i, selectedcluster[0], track, bound,impactFactor,CUDA))

                for od, each in enumerate(outs[i]):
                    if str(bound) not in self.adata.uns['%s_perturbation_deltaD'%mode].keys():
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)] = {}
                    if track_name not in self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)].keys():
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name] = {}
                    if perturbed_items[od] not in self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name].keys():
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name][perturbed_items[od]] = {}
                    if str(i) not in self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name][perturbed_items[od]].keys():
                        tempout = []
                        for kk in range(self.total_stage):
                            tempout.append(each[len(each)-self.total_stage+kk])
                        self.adata.uns['%s_perturbation_deltaD'%mode][str(bound)][track_name][perturbed_items[od]][str(i)] = tempout
    def run(self,mode,log2fc,inplace=False,random_times = 100,random_genes = 2,CUDA = True):
        
        if inplace == False:
            written=True
        else:
            written=False
        if mode == 'drug':
            for i in self.tracks.keys():

                print(i)
                self.startAutoPerturbation(i,log2fc,mode,written =written,CUDA=CUDA)
                self.startAutoPerturbation(i,1/log2fc,mode,written =written,CUDA=CUDA)
                self.hiddenReps = []
                self.perturb_stage_data_mean = []
        elif mode == 'pathway':
            for i in self.tracks.keys():
                self.startAutoPerturbation(i,log2fc,mode,written =written,CUDA=CUDA)
                self.startAutoPerturbation(i,1/log2fc,mode,written =written,CUDA=CUDA)
                self.hiddenReps = []
                self.perturb_stage_data_mean = []
        elif mode == 'random_background':
            for i in self.tracks.keys():
                self.startAutoPerturbation(i,log2fc,mode,written = written,random_times=random_times,random_genes=random_genes,CUDA=CUDA)
                self.startAutoPerturbation(i,1/log2fc,mode, written = written,random_times=random_times,random_genes=random_genes,CUDA=CUDA)
                self.hiddenReps = []
                self.perturb_stage_data_mean = []
        elif mode == 'online_random_background':
            for i in self.tracks.keys():
                self.startAutoPerturbation(i,1,mode,random_times=random_times,written =written,CUDA=CUDA)
        elif mode == 'perfect':
            for i in self.tracks.keys():

                print(i)
                self.startAutoPerturbation(i,log2fc,mode,written =written,CUDA=CUDA)
                self.startAutoPerturbation(i,1/log2fc,mode,written =written,CUDA=CUDA)
                self.hiddenReps = []
                self.perturb_stage_data_mean = []
    def run_online_speed(self, allTracks:bool,perturbated_gene,perturbated_gene_reversed, unit_name,stage = None, lastCluster=None,CUDA=False):
        '''
        allTracks: one track or all tracks bool
        stage: stage to be perturbed
        lastCluster: last cluster to be perturbed (if allTracks is False)
        perturbated_gene: gene to be perturbed format a:0.5, b: 2.5, c:0.5...
        perturbated_gene_reversed: gene to be perturbed format a:2.0, b: 0.4, c:2.0... (reversed log2fc to the original)
        unit_name: name of the unit to be perturbed
        stage: stage to be perturbed, if None choose all
        CUDA: whether to use CUDA
        '''
        import time

        online_analyst = perturbationAnalysis(self.adata,self.idrem_dir,stage=stage,allTracks=allTracks)
        # else:

            # online_analyst = perturbationAnalysis(self.adata,os.path.join(self.target_directory,'idremVizCluster'),allTracks=allTracks)
        perturbated_gene = perturbated_gene.split(',')
        perturbated_gene_reversed = perturbated_gene_reversed.split(',')
        if allTracks != True:
            out1 = {}
            out2 = {}
            track = self.getTrack(len(self.stageadata)-1,lastCluster)
            track_name = str(track[0][0])
            for i in range(1,len(track)):
                track_name += '-' + str(track[i][0])
            
            out1[track_name] = self.startAutoPerturbation_online(lastCluster,perturbated_gene,CUDA=CUDA)
            out2[track_name] = self.startAutoPerturbation_online(lastCluster,perturbated_gene_reversed,CUDA=CUDA)
            track = self.getTrack(len(self.stageadata)-1,lastCluster)

            perturbation_score, pval,out_deltaD = online_analyst.online_analysis([track_name,[out1,out2]])
            self.hiddenReps = []
            self.perturb_stage_data_mean = []
        else:
            out1 = {}
            out2 = {}
            last_clusters = []
            track_names = []
            for i in self.tracks.keys():
            
                track = self.getTrack(len(self.stageadata)-1,i)
                last_clusters.append(track[-1][0])
                track_name = str(track[0][0]) 
                for i in range(1,len(track)):
                    track_name += '-' + str(track[i][0])
                track_names.append(track_name)
            
            out1,out2 = self.startAutoPerturbation_online_speed(last_clusters,[perturbated_gene,perturbated_gene_reversed],track_names,CUDA=CUDA)
            # out2 = self.startAutoPerturbation_online_speed(last_clusters,perturbated_gene_reversed,track_names,CUDA=CUDA)
               
            step6_start = time.time()
            perturbation_score, pval,out_deltaD = online_analyst.online_analysis([out1,out2])
            step6_end = time.time()
            print('step6 time: ', step6_end - step6_start)
        return perturbation_score, pval,out_deltaD
    def run_online(self, allTracks:bool,perturbated_gene,perturbated_gene_reversed, unit_name,stage = None, lastCluster=None,CUDA=False):
        '''
        allTracks: one track or all tracks bool
        stage: stage to be perturbed
        lastCluster: last cluster to be perturbed (if allTracks is False)
        perturbated_gene: gene to be perturbed format a:0.5, b: 2.5, c:0.5...
        perturbated_gene_reversed: gene to be perturbed format a:2.0, b: 0.4, c:2.0... (reversed log2fc to the original)
        unit_name: name of the unit to be perturbed
        stage: stage to be perturbed, if None choose all
        CUDA: whether to use CUDA
        '''
        import time

        online_analyst = perturbationAnalysis(self.adata,self.idrem_dir,stage=stage,allTracks=allTracks)
        # else:

            # online_analyst = perturbationAnalysis(self.adata,os.path.join(self.target_directory,'idremVizCluster'),allTracks=allTracks)
        perturbated_gene = perturbated_gene.split(',')
        perturbated_gene_reversed = perturbated_gene_reversed.split(',')
        if allTracks != True:
            out1 = {}
            out2 = {}
            track = self.getTrack(len(self.stageadata)-1,lastCluster)
            track_name = str(track[0][0])
            for i in range(1,len(track)):
                track_name += '-' + str(track[i][0])
            
            out1[track_name] = self.startAutoPerturbation_online(lastCluster,perturbated_gene,CUDA=CUDA)
            out2[track_name] = self.startAutoPerturbation_online(lastCluster,perturbated_gene_reversed,CUDA=CUDA)
            track = self.getTrack(len(self.stageadata)-1,lastCluster)

            perturbation_score, pval,out_deltaD = online_analyst.online_analysis([track_name,[out1,out2]])
            self.hiddenReps = []
            self.perturb_stage_data_mean = []
        else:
            out1 = {}
            out2 = {}
            for i in self.tracks.keys():
            
                track = self.getTrack(len(self.stageadata)-1,i)
                track_name = str(track[0][0])
                for j in range(1,len(track)):
                    track_name += '-' + str(track[j][0])
         
                out1= self.startAutoPerturbation_online(track[-1][0],perturbated_gene,CUDA=CUDA)
                out2 = self.startAutoPerturbation_online(track[-1][0],perturbated_gene_reversed,CUDA=CUDA)
                self.hiddenReps = []
                self.perturb_stage_data_mean = []
            step6_start = time.time()
            perturbation_score, pval,out_deltaD = online_analyst.online_analysis([out1,out2])
            step6_end = time.time()
            print('step6 time: ', step6_end - step6_start)
        return perturbation_score, pval,out_deltaD
    def analysis(self,mode,log2fc,all=True,stage=None):
        # self.adata = self.read_mergeadata()
        self.adata.obs['leiden'] = self.adata.obs['leiden'].astype('string')
        self.adata.obs['stage'] = self.adata.obs['stage'].astype('string')
        analyst = perturbationAnalysis(self.adata,self.idrem_dir, stage=stage,log2fc = log2fc, mode = mode)
       
        
        temp = analyst.main_analysis(track_to_analysis = 'all', all=all, score='avg_backScore', items=None)#read item from disk for now
        if '%s_perturbation_score'%mode not in self.adata.uns.keys():
            self.adata.uns['%s_perturbation_score'%mode] = {}
        
        self.adata.uns['%s_perturbation_score'%mode][str(log2fc)] = temp
        self.adata.uns['%s_perturbation_score'%mode][str(1/log2fc)] = temp
        


#drug analysis
#load results csv which are distants.
import csv
import json
import os
import scanpy as sc
import numpy as np
import pandas as pd
from scipy.stats import norm
class perturbationAnalysis:
    def __init__(self,adata,target_directory,stage = None, log2fc=None,mode=None,allTracks=None):
        self.adata = adata
        self.log2fc = log2fc
        self.mode = mode
        self.target_directory = target_directory
        self.stage = stage
        self.allTracks = allTracks
        self.idrem_path = self.target_directory#os.path.join(self.target_directory,'idremVizCluster/')#'./'+ self.target_directory +'/idremVizCluster/'
    def readIdremJson(self, filename):
        # print('getting Target genes from ', filename)
        path = os.path.join(self.idrem_path,filename ,'DREM.json')
        f=open(path,"r")
        lf=f.readlines()
        f.close()
        lf="".join(lf)
        lf=lf[5:-2]+']'
        tt=json.loads(lf,strict=False)
        return tt
    def getTendency(self,filename):
        '''
        get the tendency of each path
        args:
        path: the file path of IDREM results
        
        return:
        out: a list of tendency of each path
        '''
        out = []
        tt = self.readIdremJson(filename)
        temp = np.array(tt[8])
        idrem_genes = temp[1:,0].tolist()
        tendency = (temp[1:,-1].astype(float)- temp[1:,1].astype(float))/3
        # stage_expression = temp[1:,1:5].astype(float)
        return tendency

    def getTendencyFromIDREM(self):
        '''
        get the tendency of each path
        args:
        path: the file path of IDREM results
        
        return:
        out: a list of tendency of each path
        '''
        out = {}
        filenames = os.listdir(self.idrem_path)
        for each in filenames:
            name = each
            if each[0] != '.':
                each = each.split('.')[0].split('-')[-1].split('n')
                for each1 in each:
                    out[each1] = self.getTendency(name)
                
        return out
    def get_tracks(self):
        # dirctory = dirctory+'_0.5.csv'
        # k = open(dirctory)#open('./iterativeTrainingNOV26/4/perturbation0.5ImportantTrack.csv')
        # kk = csv.reader(k,delimiter=',')
        if 'online_random_background_perturbation_deltaD' in list(self.adata.uns.keys()):
            tracks = self.adata.uns['online_random_background_perturbation_deltaD']['A'].keys()
        else:
            tracks = self.adata.uns['%s_perturbation_deltaD'%self.mode][str(self.log2fc)].keys()
        # tracks = []
        # for each in list(deltaD.keys()):
        #     tracks.append(each)

        return list(set(tracks))
    def load_online(self,deltaD=None,sanity=False,track_percentage= None):
        out = {}
        if sanity == True:
            k1 = self.adata.uns['online_random_background_perturbation_deltaD']['A'] #track name, times, i
            k2 = self.adata.uns['online_random_background_perturbation_deltaD']['B']
        else:
            k1 = deltaD[0]#{trackname, 'online', 'i'}
            k2 = deltaD[1]
        pathwaydic1= {}

        for each in list(k1.keys()):

            if each not in pathwaydic1.keys():
                
                for each1 in list(k1[each].keys()):
                    if each1 not in pathwaydic1.keys():
                        pathwaydic1[each1] = {}
                    
                    if each not in pathwaydic1[each1].keys():
                        pathwaydic1[each1][each] = []
                    total_len = len(k1[each][each1])
                    if track_percentage is not None:
                        pathwaydic1[each1][each] = [np.array(k1[each][each1][str(i)]) for i in range(total_len)]*np.array(track_percentage[each])
                    else:
                        pathwaydic1[each1][each] = [np.array(k1[each][each1][str(i)]) for i in range(total_len)]
        pathwaydic2= {}
        for each in list(k2.keys()):

            if each not in pathwaydic2.keys():
                
                for each1 in list(k2[each].keys()):
                    if each1 not in pathwaydic2.keys():
                        pathwaydic2[each1] = {}
                    
                    if each not in pathwaydic2[each1].keys():
                        pathwaydic2[each1][each] = []
                    total_len = len(k2[each][each1])
                    if track_percentage is not None:
                        pathwaydic2[each1][each] = [np.array(k2[each][each1][str(i)]) for i in range(total_len)]*np.array(track_percentage[each]) #each is track, each1 is item name
                    else:
                        pathwaydic2[each1][each] = [np.array(k2[each][each1][str(i)]) for i in range(total_len)]
        if sanity == False:
            for each in list(pathwaydic1.keys()):
                temp1 = []
                temp2 = []
                out_delta = []
                for i in range(len(np.sum(np.array(list(pathwaydic1[each].values())),axis=0))):
                    if self.stage is None:
                        out_delta.append(np.sum(np.array(list(pathwaydic1[each].values())),axis=0)[i])
                        temp1.append(self.calculateScore(np.sum(np.array(list(pathwaydic1[each].values())),axis=0)[i],i))
                        temp2.append(self.calculateScore(np.sum(np.array(list(pathwaydic2[each].values())),axis=0)[i],i))
                    else:
                        if i != self.stage:
                            continue
                        else:
                            self.out_delta = np.sum(np.array(list(pathwaydic1[each].values())),axis=0)[i]
                            temp1.append(self.calculateScore(np.sum(np.array(list(pathwaydic1[each].values())),axis=0)[i],i))
                            temp2.append(self.calculateScore(np.sum(np.array(list(pathwaydic2[each].values())),axis=0)[i],i))
                temp1 = np.array(temp1)
                temp2 = np.array(temp2)
                # out[each] =(np.abs(np.mean(temp1[:,0]))+np.abs(np.mean(temp2[:,0])))/2
                out[each] =np.sqrt(((np.abs(np.mean(temp1[:,0]))+np.abs(np.mean(temp2[:,0])))/2) * ((np.abs(np.mean(temp1[:,1]))+np.abs(np.mean(temp2[:,1])))/2))
                if self.stage is None:
                    self.out_delta = np.mean(out_delta,axis=0)
                # out[each] = (np.abs(np.sum(np.array(list(pathwaydic1[each].values())),axis=0))+np.abs(np.sum(np.array(list(pathwaydic2[each].values())),axis=0)))/2 #total item-delta
        if track_percentage is None and sanity == True:
            out = {}
            
            for each in list(k1.keys()):
                out[each] = []
                for item in list(k1[each].keys()):
                    temp1 = []
                    temp2 = []
                    for i in range(len(np.sum(np.array(list(k1[each][item].values())),axis=0))):
                        if self.stage is None:
                            temp1.append(self.calculateScore(np.array(list(k1[each][item].values()))[i],i))
                            temp2.append(self.calculateScore(np.array(list(k2[each][item].values()))[i],i))
                        else:
                            if i != self.stage:
                                continue
                            else:
                                temp1.append(self.calculateScore(np.array(list(k1[each][item].values()))[i],i))
                                temp2.append(self.calculateScore(np.array(list(k2[each][item].values()))[i],i))
                    temp1 = np.array(temp1)
                    temp2 = np.array(temp2)
                    # out[each].append((np.abs(np.mean(temp1[:,0]))+np.abs(np.mean(temp2[:,0])))/2) 
                    out[each].append(np.sqrt(((np.abs(np.mean(temp1[:,0]))+np.abs(np.mean(temp2[:,0])))/2) * ((np.abs(np.mean(temp1[:,1]))+np.abs(np.mean(temp2[:,1])))/2)))
                    # out[each].append((np.abs(np.array(list(k1[each][item].values())))+np.abs(np.array(list(k2[each][item].values()))))/2)) 
        elif track_percentage is not None and sanity == True:
            out = {}
            for each in list(pathwaydic1.keys()):
                temp1 = []
                temp2 = []
                for i in range(len(np.sum(np.array(list(pathwaydic1[each].values())),axis=0))):
                    if self.stage is None:
                        temp1.append(self.calculateScore(np.sum(np.array(list(pathwaydic1[each].values())),axis=0)[i],i))
                        temp2.append(self.calculateScore(np.sum(np.array(list(pathwaydic2[each].values())),axis=0)[i],i))
                    else:
                        if i != self.stage:
                            continue
                        else:
                            temp1.append(self.calculateScore(np.sum(np.array(list(pathwaydic1[each].values())),axis=0)[i],i))
                            temp2.append(self.calculateScore(np.sum(np.array(list(pathwaydic2[each].values())),axis=0)[i],i))
                temp1 = np.array(temp1)
                temp2 = np.array(temp2)
                # out[each] = (np.abs(np.mean(temp1[:,0]))+np.abs(np.mean(temp2[:,0])))/2
                out[each] =np.sqrt(((np.abs(np.mean(temp1[:,0]))+np.abs(np.mean(temp2[:,0])))/2) * ((np.abs(np.mean(temp1[:,1]))+np.abs(np.mean(temp2[:,1])))/2))
                # out[each] = (np.abs(np.sum(np.array(list(pathwaydic1[each].values())),axis=0))+np.abs(np.sum(np.array(list(pathwaydic2[each].values())),axis=0)))/2
        return out
    def load(self,data_pathway_overlap_genes,track_percentage,all=True,sanity=False):
        if sanity == True:
            k1 = self.adata.uns['random_background_perturbation_deltaD'][str(self.log2fc)]
            k2 = self.adata.uns['random_background_perturbation_deltaD'][str(1/self.log2fc)]
        else:
            k1 = self.adata.uns['%s_perturbation_deltaD'%self.mode][str(self.log2fc)]
            k2 = self.adata.uns['%s_perturbation_deltaD'%self.mode][str(1/self.log2fc)]
        # self.stageadata[i].uns['%s_perturbation_deltaD'%mode][str(bound)][track_name][name]
        pathwaydic1= {}
        for each in list(k1.keys()):

            if each not in pathwaydic1.keys():
                
                for each1 in list(k1[each].keys()):
                    if each1 not in pathwaydic1.keys():
                        pathwaydic1[each1] = {}
                    
                    if each not in pathwaydic1[each1].keys():
                        pathwaydic1[each1][each] = []
                    total_len = len(k1[each][each1])
                    
                    if all == True:
                        pathwaydic1[each1][each] = [np.array(k1[each][each1][str(i)]) for i in range(total_len)]*np.array(track_percentage[each])
                    else:
                        pathwaydic1[each1][each] = [np.array(k1[each][each1][str(i)]) for i in range(total_len)]
       
        pathwaydic2= {}
#each 1 is the item name, each is the track name
        for each in list(k2.keys()):

            if each not in pathwaydic2.keys():
                
                for each1 in list(k2[each].keys()):
                    if each1 not in pathwaydic2.keys():
                        pathwaydic2[each1] = {}
                    
                    if each not in pathwaydic2[each1].keys():
                        pathwaydic2[each1][each] = []
                    total_len = len(k2[each][each1])
                    if all == True:
                        pathwaydic2[each1][each]= [np.array(k2[each][each1][str(i)]) for i in range(total_len)]*np.array(track_percentage[each])
                    else:
                        pathwaydic2[each1][each]= [np.array(k2[each][each1][str(i)]) for i in range(total_len)]
        if all:
            out = {}
            for each in list(pathwaydic1.keys()):
                out[each] = {}
                temp1 = []
                temp2 = []
                temp = []
                for i in range(len(np.sum(np.array(list(pathwaydic1[each].values())),axis=0))):
                    temp1.append(self.calculateScore(np.sum(np.array(list(pathwaydic1[each].values())),axis=0)[i],i))
                    temp2.append(self.calculateScore(np.sum(np.array(list(pathwaydic2[each].values())),axis=0)[i],i))
                temp1 = np.array(temp1)
                temp2 = np.array(temp2)

                temp.append(np.sqrt(((np.abs(np.mean(temp1[:,0]))+np.abs(np.mean(temp2[:,0])))/2) * ((np.abs(np.mean(temp1[:,1]))+np.abs(np.mean(temp2[:,1])))/2)))
                # temp.append((np.abs(np.mean(temp1[:,0]))+np.abs(np.mean(temp2[:,0])))/2) 
                # temp.append((np.mean(temp1[:,0])-np.mean(temp2[:,0]))/2 * ((np.abs(np.mean(temp1[:,1]))+np.abs(np.mean(temp2[:,1])))/2)) 
                out[each]['total'] = np.array(temp)
        else:
            out = {}
            max = 0
            pw = None
            tr = None
            for each in list(pathwaydic1.keys()): #each is the pathway name
                out[each] = {}
                
                for item in list(pathwaydic1[each].keys()):#item is the track name
                    temp1 = []
                    temp2 = []
                    temp = []
                    for i in range(len(np.array(list(pathwaydic1[each][item])))):
                        
                        temp1.append(self.calculateScore(np.array(list(pathwaydic1[each][item]))[i],i))
                        temp2.append(self.calculateScore(np.array(list(pathwaydic2[each][item]))[i],i))
                    temp1_copy = np.array(temp1)
                    temp2_copy = np.array(temp2)

                    # temp.append((np.abs(np.mean(temp1_copy[:,0]))+np.abs(np.mean(temp2_copy[:,0])))/2 )
                    
                    temp.append(np.sqrt(((np.abs(np.mean(temp1_copy[:,0]))+np.abs(np.mean(temp2_copy[:,0])))/2) * ((np.abs(np.mean(temp1_copy[:,1]))+np.abs(np.mean(temp2_copy[:,1])))/2)))
                    # temp.append((np.mean(temp1[:,0])-np.mean(temp2[:,0]))/2* ((np.abs(np.mean(temp1[:,1]))+np.abs(np.mean(temp2[:,1])))/2)) 
                    out[each][item] = np.array(temp)

        if sanity == False:

            for each in list(out.keys()):
                if each not in data_pathway_overlap_genes:
                    del out[each]

        return out
    
#convert distance to scores and some statistics

    def get_cluster_data_size(self,stage,cluster):

        stagedata = self.adata[self.adata.obs['stage'] == str(stage)]#.index.tolist()
        # stagedata = self.adata[stageids]
       
        clusterids = stagedata.obs[stagedata.obs['leiden'] == str(cluster)].index.tolist()
        clusterdata = stagedata[clusterids]
        return len(clusterdata)
    def get_track_percentage(self, tracks):
        percentage = {}
        import time
        stageadatas = [self.adata[self.adata.obs[self.adata.obs['stage'] == stage].index.tolist()] for stage in self.adata.obs['stage'].unique()]
        for each_track in tracks:
           
            if '-' not in each_track:
                continue
            percentage[each_track] = 0
            clusters = each_track.split('-')
        
            for stage, each_cluster in enumerate(clusters):
                
                cluster_size = stageadatas[stage].obs['leiden'].value_counts().to_dict()[each_cluster]#self.get_cluster_data_size(stageadatas[stage],each_cluster)
                percentage[each_track] += cluster_size
            percentage[each_track] /= len(self.adata)


        return percentage
    def calculateScore(self,delta,flag,weight=100):
        '''
        flag: the stage of selected clsuter
        delta: perturbation distance - simulation distance (in z space)
        
        '''
        out = 0
        out1 = 0
        separate = []
        for i, each in enumerate(delta):
            
            if i != flag:
                # print((1-1/(1+np.exp(weight*each*np.sign(i-flag)))-0.5)/0.5)
                out+=(1-1/(1+np.exp(weight*each*np.sign(i-flag)))-0.5)/0.5
                out1+=np.abs((1-1/(1+np.exp(weight*each))-0.5)/0.5)
                # separate.append((1-1/(1+np.exp(weight*each))-0.5)/0.5)
                # separate.append((1-1/(1+np.exp(weight*each*np.sign(i-flag)))-0.5)/0.5)
            # separate[0] = out/3 #just for test
            # out = separate[0]#test
   
        return out/(len(delta)-1), out1/(len(delta)-1)

    def calculateAvg(self,pathwaydic,tracklist='all',sanity=False):
        
        perturbationresultdic = {}
        perturbationresultdic['backScore'] = []
        for each in list(pathwaydic.keys()):

            perturbationresultdic[each] = {}
            

            for track in list(pathwaydic[each].keys()):
                clusterType = track.split('-')
                if tracklist != 'all':
                    target = []
                    for t in tracklist:
                        target.append(str(t))
                    #target = str(track)#['3','16']
                #target = ['3','16','4','28','23','19','9','13']
    #             if clusterType[0] not in target:
    #                 continue
                perturbationresultdic[each][track] = {}
                perturbationresultdic[each][track]['backScore'] = []
                for flag, each_perturbation in enumerate(pathwaydic[each][track]):
                    out = each_perturbation
                    perturbationresultdic[each][track]['backScore'] = out
                    
                
                perturbationresultdic[each][track]['avg_backScore'] = perturbationresultdic[each][track]['backScore']
        return perturbationresultdic



    def getStatistics(self,perturbationresultdic):
        avg_backScore = []
        backScore = []

        tempkey = list(perturbationresultdic.keys())[-1]
        key = list(perturbationresultdic.keys())
        
        # avg_backScore.pop()
        # backScore.pop()
        track_name = list(perturbationresultdic[key[-1]].keys())
        for each in track_name:
            avg_backScore.append([])
            backScore.append([])
        pathway_name_order = []
        key.remove('backScore')
        pathway_name_order = []
        for i in range(len(backScore)):
            for each in key:
                pathway_name_order.append(each)
                backScore[i].append(perturbationresultdic[each][track_name[0]]['backScore'])
                avg_backScore[i].append(perturbationresultdic[each][track_name[0]]['avg_backScore'])
                

        return  avg_backScore,  backScore,track_name,pathway_name_order

    def fitlerOutNarrowPathway(self, scores, sanity_scores, pathway_names, pathway_name_order):
        '''
        for this function, only get pathway cdf in a specific path
        
        args:
        scores: the score of all pathways in a track
        pathway_names: name of the pathway
        pathway_name_order: the pathway name order
        
        perturbationresultdic: perturbation statistic results
        pathdic: scores of pathways in every path
        sanity_pathdic: sanity (random) scores of pathways in every path
        scoreindex: name of score to print
        flag:0: pushback, 1: pushforward
        '''
        
        sanity_scores = np.array(sanity_scores)
        # print('sanity mean: ', sanity_scores.mean())
        # print('sanity std: ', sanity_scores.std())
        top_compounds = []
        down_compounds = []
        record_top = []
        record_down = []

    
        for i,each in enumerate(scores):

            if float(each) >= sanity_scores.mean()+sanity_scores.std():
                top_compounds.append(pathway_names[i])
                record_top.append(i)
            elif  float(each) <= sanity_scores.mean()-sanity_scores.std():
                
                down_compounds.append(pathway_names[i])
                record_down.append(i)
        scores = np.array(scores)
        # print(down_compounds)
        filtered_top_score = scores[record_top]
        filtered_down_score = scores[record_down]
        filtered_top_index = np.argsort(filtered_top_score).tolist()
        filtered_top_index.reverse()
        filtered_down_index = np.argsort(filtered_down_score).tolist()

        filtered_top_score = sorted(filtered_top_score,reverse=True)
        filtered_down_score = sorted(filtered_down_score)
        
        top_compounds = np.array(top_compounds)
        down_compounds = np.array(down_compounds)
        top_compounds = top_compounds[filtered_top_index]
        down_compounds = down_compounds[filtered_down_index]
        
        final_down_compounds = []
        final_top_compounds = []
        for i, each in enumerate(filtered_top_score):
            cdf = norm.cdf(each, sanity_scores.mean(),sanity_scores.std())
            # if (1.000-cdf) * len(filtered_top_score) / (i + 1)  < 0.05:
            if (1.000-cdf) <0.05:
                final_top_compounds.append(top_compounds[i])
        for i, each in enumerate(filtered_down_score):
            cdf = norm.cdf(each, sanity_scores.mean(),sanity_scores.std())
            if cdf * len(filtered_down_score) / (i + 1) < 0.05:
                # print('down compound:',down_compounds[i])
                final_down_compounds.append(down_compounds[i])
        return top_compounds,down_compounds
    def getTopDownPathways(self,pathwaydic,track_percentage,track,all=all, flag=0):
        '''
        flag: 0:both pushback score and pushforward score; 1 only pushback score, -1 only pushforward score
        
        '''

        perturbationresultdic = self.calculateAvg(pathwaydic,tracklist=track)
        sanity_pathwaydic = self.load( None ,track_percentage,all=all,sanity=True)
        sanity_perturbationresultdic = self.calculateAvg(sanity_pathwaydic,sanity=True,tracklist=track)
        

        pathways = list(perturbationresultdic.keys())
        pathways.remove('backScore')

        # pathdic,sanity_pathdic = conver_pathway_track_score_to_track_pathway_score(pathways,perturbationresultdic,sanity_perturbationresultdic,scoreindex)
        sanity_avg_backScore, sanity_backScore,sanity_track_name,_ = self.getStatistics(sanity_perturbationresultdic)

        avg_backScore, backScore,track_name,pathway_name_order = self.getStatistics(perturbationresultdic)

        outtext=[]
        results= {}
        for i, each in enumerate(avg_backScore):
            
            results[track_name[i]] = {}
            
            top_compounds,down_compounds=self.fitlerOutNarrowPathway(each,sanity_avg_backScore[i],pathways,pathway_name_order)
            # print(down_compounds)
            results[track_name[i]]['top_compounds']={}
            results[track_name[i]]['down_compounds']= {}
            for j,each in enumerate(top_compounds):
                results[track_name[i]]['top_compounds'][str(j)] = top_compounds[j]
            for j,each in enumerate(down_compounds):  
                results[track_name[i]]['down_compounds'][str(j)] = down_compounds[j]
        results = pd.json_normalize(results)
        results.columns = results.columns.str.split(".").map(tuple)
        results = results.stack([0, 1]).reset_index(0, drop=True)
        results = results.transpose()
        # return results
        results = results.reset_index()

        results['index'] = results['index'].astype('int32')
        results = results.set_index(['index'])
        results =results.sort_index()
    
        return results

    def conver_pathway_track_score_to_track_pathway_score(self,pathway,perturbationresultdic,sanity_perturbationresultdic,score,track):
        '''
        previously the dict is pathway-track-scores, convert it to track-scores
        
        args:
        pathway: a list of names of pathways
        perturbationresultdic: perturbation statistic results
        '''
        
        pathdic = {}
        sanity_pathdic = {}
        for path in list(perturbationresultdic[pathway[0]].keys()):

            if path == 'statistic':
                continue
            if track == 'all':
                pathdic[path] = []
                sanity_pathdic[path]=[]
            else:
                if path.split('-')[0] in track:
                    pathdic[path] = []
                    sanity_pathdic[path]=[]
        
        for each in list(perturbationresultdic.keys()):
            if each != 'backScore':

                for k in list(perturbationresultdic[each].keys()):
                    if k!= 'statistic' and k in pathdic.keys():
                        # pathdic[k].append(perturbationresultdic[each][k][score][0])
                        pathdic[k].append(perturbationresultdic[each][k][score]) #avg score
        for each in list(sanity_perturbationresultdic.keys()):
            if each != 'backScore':
                for k in list(sanity_perturbationresultdic[each].keys()):
                    if k!= 'statistic' and k in sanity_pathdic.keys():
                        # sanity_pathdic[k].append(sanity_perturbationresultdic[each][k][score][0])
                        sanity_pathdic[k].append(sanity_perturbationresultdic[each][k][score])#avg score
    
        return pathdic,sanity_pathdic


    def getTrackPathwayCDF(self,pathway,perturbationresultdic,path,pathdic,sanity_pathdic,gene_in_pathway,scoreindex,flag=0):
        '''
        for this function, only get pathway cdf in a specific path
        
        args:
        pathway: the pathway to get cdf
        path: path to get pathway cdf 
        perturbationresultdic: perturbation statistic results
        pathdic: scores of pathways in every path
        sanity_pathdic: sanity (random) scores of pathways in every path
        scoreindex: name of score to print
        flag:0: pushback, 1: pushforward
        '''
    
        pathdic[path] = np.array(pathdic[path])
        sanity_pathdic[path]= np.array(sanity_pathdic[path])
    
        score = perturbationresultdic[pathway][path][scoreindex]

        
        cdf=norm.cdf(perturbationresultdic[pathway][path][scoreindex],sanity_pathdic[path].mean(),sanity_pathdic[path].std())    
        return score, 1.0000000-cdf
    def getSummarizedResults(self,track_to_analysis,pathwayranking, pathwaydic,track_percentage,gene_in_pathway,scoreindex,direction_dict,all):#(pathwayranking,perturbationresultdic,pathdic,sanity_pathdic,gene_in_pathway,score):
        '''
        args:
        pathwayranking: ordered pathway based on scores (both top and down)
        pathwaydic: pathway dictionary, distance info
        gene_len_dict: number of gene in each pathway
        sanity_directory: directory to sanity results
        gene_in_pathway:  genes in each pathway overlapping with adatagene
        scoreindex: perturbation score
        
        '''

        sanity_pathwaydic = self.load(gene_in_pathway,track_percentage,all=all,sanity=True)
        # np.save('sanity_pathwaydic.npy',sanity_pathwaydic)
        # np.save('pathwaydic.npy',pathwaydic)
        # print('saved')
        # print(gdsg)
        sanity_perturbationresultdic = self.calculateAvg(sanity_pathwaydic,sanity=True,tracklist=track_to_analysis)
        perturbationresultdic = self.calculateAvg(pathwaydic,tracklist=track_to_analysis)
        pathways = list(pathwaydic.keys())
        if all:
            pathdic,sanity_pathdic = self.conver_pathway_track_score_to_track_pathway_score(pathways,perturbationresultdic,sanity_perturbationresultdic,scoreindex,'total')
        else:
            pathdic,sanity_pathdic = self.conver_pathway_track_score_to_track_pathway_score(pathways,perturbationresultdic,sanity_perturbationresultdic,scoreindex,track_to_analysis)
        tendency_dict = self.getTendencyFromIDREM()

        tendency = []
        for each in list(track_percentage.keys()):
            each1 = each.split('-')[-1]
            if len(tendency) == 0:
                tendency = np.array(tendency_dict[each1])*track_percentage[each]
            else:
                tendency = tendency + np.array(tendency_dict[each1])*track_percentage[each]
        infodict = {}
        pathwayranking = dict(pathwayranking)
        for outer, inner in pathwayranking.keys():
            if outer not in infodict.keys():
                infodict[outer] = {}
            infodict[outer][inner] = {}
            infodict[outer][inner]['compound'] = {}
            for idx,each in enumerate(list(pathwayranking[(outer,inner)])):
                infodict[outer][inner]['compound'][str(idx)]=each

        for track in list(infodict.keys()):
            # del infodict[track]['down_compounds']
            for updown in list(infodict[track].keys()):

                infodict[track][updown]['perturbation score'] = {}
                infodict[track][updown]['pval_adjusted'] = {}
                infodict[track][updown]['drug_regulation'] = {}
                infodict[track][updown]['idrem_suggestion'] = {}
                count = 0
                for idx in range(len(infodict[track][updown]['compound'].keys())):
                    eachpathway = infodict[track][updown]['compound'][str(idx)]
                    if eachpathway == 'backScore':
                        continue
                    if str(eachpathway) == 'nan':
                        continue
                    count+=1
                
                for idx in range(len(infodict[track][updown]['compound'].keys())):
                    eachpathway = infodict[track][updown]['compound'][str(idx)]

                    if eachpathway == 'backScore':
                        continue
                    if str(eachpathway) == 'nan':
                        continue
                    flag = 0
                    if updown == 'down_compounds':
                        flag = 1
                
                    score, p=self.getTrackPathwayCDF(eachpathway,perturbationresultdic,track,pathdic, sanity_pathdic, gene_in_pathway, scoreindex,flag=flag)
                    p=p*count/(count-idx)
                    if p >1:
                        p = 1
                    infodict[track][updown]['perturbation score'][str(idx)] =score
                    infodict[track][updown]['pval_adjusted'][str(idx)] = p
                    infodict[track][updown]['drug_regulation'][str(idx)] = gene_in_pathway[eachpathway]
                    written_gene = []
                    for eachgene in gene_in_pathway[eachpathway]:
                        eachgene = eachgene.split(':')[0]
                        temp = direction_dict[eachgene]
                        if temp > 0: #increasing tendency decrease gene expression
                            written_gene.append(eachgene+':-')
                        else: #decreasing tendency increase gene expression
                            written_gene.append(eachgene+':+')
                    infodict[track][updown]['idrem_suggestion'][str(idx)] = written_gene
        # infodict = pd.json_normalize(infodict)
        # infodict.columns = infodict.columns.str.split(".").map(tuple)
        # infodict = infodict.stack([0, 1, 2]).reset_index(0, drop=True)
        # infodict = infodict.transpose()
        # infodict = infodict.reset_index()
        # infodict['index'] = infodict['index'].astype('int32')
        # infodict = infodict.set_index(['index']).sort_index()
        return infodict
    def getTendencyDict(self,track_percentage):
        genenames = self.adata.var.index.tolist()
        tendency_dict = self.getTendencyFromIDREM()
        tendency = None
        for each in list(track_percentage.keys()):
            each1 = each.split('-')[-1]
            if tendency is None:
                tendency = np.array(tendency_dict[each1])*track_percentage[each]
            else:
                tendency = tendency + np.array(tendency_dict[each1])*track_percentage[each]
        output = {}
        for i, each in enumerate(genenames):
            if  each not in output.keys():
                output[each] = tendency[i]
        return output
    def get_online_results(self, score, random_score):
        cdf=norm.cdf(score,np.array(random_score).mean(),np.array(random_score).std())
        return 1.00 - cdf
#load avg needed

    def main_analysis(self,track_to_analysis, all,score=None,items=None,custommode=None):
        '''
        all: whether to put all tracks together
        '''

        tracks = self.get_tracks()
        track_percentage = self.get_track_percentage(tracks)
        if self.mode == 'pathway':
            items = self.adata.uns['data_pathway_overlap_genes']
            #  items=dict(np.load('./iterativeTrainingNOV26/4/data_pathway_overlap_genes.npy',allow_pickle=True).tolist())#together up and down
        else:
            items = self.adata.uns['data_drug_overlap_genes']
            #=dict(np.load('./iterativeTrainingNOV26/4/dec23_drug_target_with_direction.npy',allow_pickle=True).tolist())
        # if items is None:
        #     items=dict(np.load('./iterativeTrainingNOV26/4/dec23_drug_target_with_direction.npy',allow_pickle=True).tolist())#together up and down
        # else:
        #     items=dict(np.load(items,allow_pickle=True).tolist())
        direction_dict = self.getTendencyDict(track_percentage)
        pathwaydic = self.load(items,track_percentage,all=all)
        
        topdownpathways = self.getTopDownPathways(pathwaydic,track_percentage,track_to_analysis,all=all)
        return self.getSummarizedResults(track_to_analysis,topdownpathways,pathwaydic,track_percentage,items,score,direction_dict,all=all)
    def online_analysis(self,deltaD):
        import time

        self.adata.obs['stage'] = self.adata.obs['stage'].astype('string')
        if self.allTracks == True:
            tracks = self.get_tracks()
            
            track_percentage = self.get_track_percentage(tracks)
            
            
            perturbation_scores = self.load_online(deltaD,track_percentage=track_percentage)
            
            random_scores = self.load_online(sanity=True,track_percentage=track_percentage)
            
        elif self.allTracks == False:
            track = deltaD[0]
            perturbation_scores =  self.load_online(deltaD[1]) # if one track mode, delta = [track,[deltaDA,deltaDB]]
            random_scores =self.load_online(sanity=True)
        
       
        perturbation_score = np.array(list(perturbation_scores.values()))#.mean()
        random_score = np.array(list(random_scores.values()))
        
        pval = self.get_online_results(perturbation_score, random_score)
        
        
        print(perturbation_score, pval, self.out_delta)
        
        return perturbation_score, pval, self.out_delta



from pyro import poutine
import pyro.ops.jit
from torch.nn.parameter import Parameter
from pyro.infer.util import is_validation_enabled
from pyro.infer.enum import get_importance_trace
from pyro.poutine.util import prune_subsample_sites
from pyro.distributions.util import is_identically_zero, scale_and_mask
from pyro.infer.elbo import ELBO
from pyro.util import check_model_guide_match, check_site_shape, ignore_jit_warnings, warn_if_nan, warn_if_inf
from pyro.infer.util import (
    MultiFrameTensor,
    get_plate_stacks,
    is_validation_enabled,
    torch_item,
)
import torch.nn.functional as F
import torch.nn as nn
import torch
from pyro.util import check_if_enumerated, warn_if_nan

def graphUpdater(loss, model, guide,discriminator, optim,x,adj,i, start,end,device='cuda:0',second_optimizer=None,two=False):
    '''
    new gcn updater, try to tackle big batch size issue
    '''
    
    with poutine.trace(param_only=True) as param_capture:
        loss_vae,surrogate_loss_particle,loss_discriminator = loss.loss_and_grads(model, guide,discriminator, device, x, adj,i, start,end)
        surrogate_loss_particle.backward()
    params = set(site["value"].unconstrained() for site in param_capture.trace.nodes.values()  if 'discriminator' not in site['name']  and 'recon' not in site['name'])
    optim(params)
    
    # zero gradients
    pyro.infer.util.zero_grads(params)
   
    if second_optimizer is not None:
        optim = second_optimizer
    with poutine.trace(param_only=True) as param_capture:
        loss_vae,surrogate_loss_particle,loss_discriminator = loss.loss_and_grads(model, guide,discriminator, device, x, adj,i, start,end)
    params = set(site["value"].unconstrained() for site in param_capture.trace.nodes.values() if 'discriminator' in site['name'] and 'recon' not in site['name'])
    loss_discriminator.backward()
    optim(params)
    # zero gradients
    pyro.infer.util.zero_grads(params)
    loss = loss_vae
    loss+=loss_discriminator
    #print(loss_discriminator)
#     loss = loss_vae
    if two == True:
        return torch_item(loss_vae), torch_item(loss_discriminator)
    if isinstance(loss, tuple):
        # Support losses that return a tuple, e.g. ReweightedWakeSleep.
        return type(loss)(map(torch_item, loss))
    else:
        return torch_item(loss)
def tuneGraphUpdater(loss, model, guide,discriminator, optim,x,adj,i, start,end,device='cuda:1',second_optimizer=None,two=False):
    '''
    tune with pushback scores
    '''
    
    with poutine.trace(param_only=True) as param_capture:
        loss_vae,surrogate_loss_particle,loss_discriminator = loss.loss_and_grads(model, guide,discriminator, device, x, adj,i, start,end)
        surrogate_loss_particle.backward()
    params = set(site["value"].unconstrained() for site in param_capture.trace.nodes.values()  if 'discriminator' not in site['name']  and 'recon' not in site['name'])
    optim(params)
    
    # zero gradients
    pyro.infer.util.zero_grads(params)
   
    if second_optimizer is not None:
        optim = second_optimizer
    with poutine.trace(param_only=True) as param_capture:
        loss_vae,surrogate_loss_particle,loss_discriminator = loss.loss_and_grads(model, guide,discriminator, device, x, adj,i, start,end)
    params = set(site["value"].unconstrained() for site in param_capture.trace.nodes.values() if 'discriminator' in site['name'] and 'recon' not in site['name'])
    loss_discriminator.backward()
    optim(params)
    # zero gradients
    pyro.infer.util.zero_grads(params)
    loss = surrogate_loss_particle
    loss+=loss_discriminator
    #print(loss_discriminator)
#     loss = loss_vae
    if two == True:
        return torch_item(loss_vae), torch_item(loss_discriminator)
    if isinstance(loss, tuple):
        # Support losses that return a tuple, e.g. ReweightedWakeSleep.
        return type(loss)(map(torch_item, loss))
    else:
        return torch_item(loss)
    
def mySigmoid(z):
    '''
    shifted sigmoid transformation of given data
    args: 
    z: input data
    return:  
    out: data after shifted sigmoid transformation
    '''
    z = np.array(z)
    out = 2*1/(1+np.exp(-z))
    return out
class myELBO(ELBO):
    def __init__(self, geneWeight=None, pushback_Score = None):
    
        super(myELBO, self).__init__()
        self.geneWeight = geneWeight
        self.pushback_Score = pushback_Score
    def get_importance_trace(self,
    graph_type, max_plate_nesting, model, guide,discriminator, args, kwargs,detach=False
):
        """
        Returns a single trace from the guide, which can optionally be detached,
        and the model that is run against it.
        """
        pyro.module("discriminator", discriminator)
        
        guide_trace = poutine.trace(guide, graph_type=graph_type).get_trace(*args, **kwargs)
        if detach:
            guide_trace.detach_()
        model_trace = poutine.trace(poutine.replay(model, trace=guide_trace), graph_type=graph_type).get_trace(*args, **kwargs)
        if is_validation_enabled():
            check_model_guide_match(model_trace, guide_trace, max_plate_nesting)

        guide_trace = prune_subsample_sites(guide_trace)
        model_trace = prune_subsample_sites(model_trace)

        model_trace = self.compute_log_prob(model_trace)
        guide_trace.compute_score_parts()

            
        if is_validation_enabled():
            for site in model_trace.nodes.values():
                if site["type"] == "sample":
                    check_site_shape(site, max_plate_nesting)
            for site in guide_trace.nodes.values():
                if site["type"] == "sample":
                    check_site_shape(site, max_plate_nesting)

        return model_trace, guide_trace
    def _get_traces(self, model, guide,discriminator, args, kwargs):
        """
        Runs the guide and runs the model against the guide with
        the result packaged as a trace generator.
        """
        if self.vectorize_particles:
            if self.max_plate_nesting == float("inf"):
                self._guess_max_plate_nesting(model, guide,discriminator, args, kwargs)
            yield self._get_vectorized_trace(model, guide,discriminator, args, kwargs)
        else:
            for i in range(self.num_particles):
                yield self._get_trace(model, guide,discriminator, args, kwargs)
    def _get_trace(self, model, guide,discriminator, args, kwargs):
        """
        Returns a single trace from the guide, and the model that is run
        against it.
        """
        model_trace, guide_trace = self.get_importance_trace(
            "flat", self.max_plate_nesting, model, guide,discriminator, args, kwargs
        )
        if is_validation_enabled():
            check_if_enumerated(guide_trace)
        return model_trace, guide_trace
    def differentiable_loss(self, model, guide, *args, **kwargs):
        """
        Computes the surrogate loss that can be differentiated with autograd
        to produce gradient estimates for the model and guide parameters
        """
        loss = 0.0
        surrogate_loss = 0.0
        for model_trace, guide_trace in self._get_traces(model, guide, args, kwargs):
            loss_particle, surrogate_loss_particle = self._differentiable_loss_particle(
                model_trace, guide_trace
            )
            surrogate_loss += surrogate_loss_particle / self.num_particles
            loss += loss_particle / self.num_particles
        warn_if_nan(surrogate_loss, "loss")
        return loss + (surrogate_loss - torch_item(surrogate_loss))

    def loss(self, model, guide, *args, **kwargs):
        """
        :returns: returns an estimate of the ELBO
        :rtype: float

        Evaluates the ELBO with an estimator that uses num_particles many samples/particles.
        """
        elbo = 0.0
        
        for model_trace, guide_trace in self._get_traces(model, guide, args, kwargs):
            elbo_particle = torch_item(my_log_sum(model_trace)) - torch_item(my_log_sum(guide_trace))
            elbo += elbo_particle / self.num_particles
        loss = -elbo
        warn_if_nan(loss, "loss")
        return loss
    def _differentiable_loss_particle(self, model_trace, guide_trace):
        elbo_particle = 0
        surrogate_elbo_particle = 0
        log_r = None
        # compute elbo and surrogate elbo
        for name, site in model_trace.nodes.items():
            if site["type"] == "sample":
                elbo_particle = elbo_particle + torch_item(site["log_prob_sum"])
                surrogate_elbo_particle = surrogate_elbo_particle + site["log_prob_sum"]
        
        for name, site in guide_trace.nodes.items():
            if site["type"] == "sample":
                
                log_prob, score_function_term, entropy_term = site["score_parts"]

                elbo_particle = elbo_particle - torch_item(site["log_prob_sum"])

                if not is_identically_zero(entropy_term):
                    surrogate_elbo_particle = (
                        surrogate_elbo_particle - entropy_term.sum()
                    )

                if not is_identically_zero(score_function_term):
                    if log_r is None:
                        log_r = _compute_log_r(model_trace, guide_trace)
                    site = log_r.sum_to(site["cond_indep_stack"])
                    surrogate_elbo_particle = (
                        surrogate_elbo_particle + (site * score_function_term).sum()
                    )

        return -elbo_particle, -surrogate_elbo_particle
    def _sum_rightmost(self,value, dim):
        r"""
        Sum out ``dim`` many rightmost dimensions of a given tensor.

        Args:
            value (Tensor): A tensor of ``.dim()`` at least ``dim``.
            dim (int): The number of rightmost dims to sum out.
        """
        if dim == 0:
            return value
        required_shape = value.shape[:-dim] + (-1,)
        return value.reshape(required_shape).sum(-1)
    def loss_and_grads(self, model, guide, discriminator, device, *args, **kwargs):
        """
        :returns: returns an estimate of the ELBO
        :rtype: float

        Computes the ELBO as well as the surrogate ELBO that is used to form the gradient estimator.
        Performs backward on the latter. Num_particle many samples are used to form the estimators.
        """
        loss = 0.0
        
        # grab a trace from the generator
        for model_trace, guide_trace in self._get_traces(model, guide,discriminator, args, kwargs):
            
#          
            
            loss_particle, surrogate_loss_particle = self._differentiable_loss_particle(
                model_trace, guide_trace
            )
#             print(model_trace.nodes['obs']['value'])
#             print(model_trace.nodes['recon']['value'].detach())

            x_true = model_trace.nodes['obs']['value'].to(device)
            y_true = torch.ones(x_true.shape[0],1).float().to(device)
            x_fake = model_trace.nodes['recon']['value'].to(device)
            y_fake = torch.zeros(x_true.shape[0],1).float().to(device)
            loss_discriminator = discriminator(x_true,y_true)
#             print('discriminator_true', loss_discriminator)
            discriminator_fake = discriminator(x_fake,y_fake)
#             print('discriminator_fake', discriminator_fake)
            loss_discriminator += discriminator_fake
            loss_discriminator = (loss_discriminator / self.num_particles)
            vae_add = discriminator(x_fake,y_true)
#             print('vae_add', vae_add)
            
            loss_generator = surrogate_loss_particle + vae_add
            #surrogate_loss_particle+=vae_add
#             print(sum(x)/len(x))
#             print(x)
#             print(dgdg)
#             loss_discriminator = loss_discriminator_fn()
#             print(loss_discriminator)
#             print(loss_particle)
#             discriminator(model_trace.nodes['obs'],)
            
            if self.pushback_Score is not None:
                pushback_Score = Parameter(torch.sum(self.pushback_Score))

                loss += 0.00001*(loss_particle) / self.num_particles + vae_add + 10000*pushback_Score#self.pushback_Score
            
            else:
                loss += (loss_particle) / self.num_particles + vae_add
            
            # collect parameters to train from model and guide
            trainable_params = any(
                site["type"] == "param"
                for trace in (model_trace, guide_trace)
                for site in trace.nodes.values()
                
            )
           
            if trainable_params and getattr(
                surrogate_loss_particle, "requires_grad", False
            ):
                if self.pushback_Score is not None:
                    surrogate_loss_particle = 0.00001* surrogate_loss_particle / self.num_particles + vae_add + 10000*pushback_Score
                else:
                    surrogate_loss_particle = surrogate_loss_particle / self.num_particles + vae_add
        warn_if_nan(loss, "loss")
        return loss,surrogate_loss_particle,loss_discriminator
    def compute_log_prob(self,trace, site_filter=lambda name, site: True):
        """
        Compute the site-wise log probabilities of the trace.
        Each ``log_prob`` has shape equal to the corresponding ``batch_shape``.
        Each ``log_prob_sum`` is a scalar.
        Both computations are memoized.
        """
        
        for name, site in trace.nodes.items():
            if site["type"] == "sample" and site_filter(name, site):
                if "log_prob" not in site:
                    try:
                        if self.geneWeight is not None and name == 'obs':
                            #print(site["value"].shape)
                            #log_p_gene = site["fn"].base_dist.log_prob(site["value"])*self.geneWeight
                            # print(site["fn"].log_prob(site["value"]))
                            
                            log_p_gene = site["fn"].base_dist.log_prob(site["value"])*self.geneWeight
                            log_p = self._sum_rightmost(log_p_gene,1)#1 is the indicator of event, showed in model.sample.event
                            #log_p = site["fn"].log_prob(site["value"], *site["args"], **site["kwargs"])
                        else:
                            log_p = site["fn"].log_prob(
                            site["value"], *site["args"], **site["kwargs"])
                            # log_p = log_p.mean()
                        #print(log_p.shape)
                        
                    except ValueError as e:
                        _, exc_value, traceback = sys.exc_info()
                        shapes = self.format_shapes(last_site=site["name"])
                        raise ValueError(
                            "Error while computing log_prob at site '{}':\n{}\n{}".format(
                                name, exc_value, shapes
                            )
                        ).with_traceback(traceback) from e
                    site["unscaled_log_prob"] = log_p
                    log_p = scale_and_mask(log_p, site["scale"], site["mask"])
                    site["log_prob"] = log_p
                    site["log_prob_sum"] = log_p.sum()
                    # site["log_prob_sum"] = log_p.mean() #use mean to try to tackle large batch size nan issue
                    if is_validation_enabled():
                        warn_if_nan(
                            site["log_prob_sum"],
                            "log_prob_sum at site '{}'".format(name),
                        )
                        warn_if_inf(
                            site["log_prob_sum"],
                            "log_prob_sum at site '{}'".format(name),
                            allow_neginf=True,
                        )
        return trace


from .plot_cell_embeddings import plot_stages_latent_representation
from .plot_cell_type_composition import plot_cell_type_composition
from .plot_hierarchical_markers_heatmap import hierarchical_static_markers_heatmap
from .plot_stage_dendrogram import plot_hc_dendrogram


import numpy as np
import gc
import os
import anndata
import pandas as pd
import scanpy as sc
from sklearn.metrics import silhouette_score, davies_bouldin_score
from sklearn.neighbors import KernelDensity
from sklearn import cluster
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
from sklearn.mixture import GaussianMixture
from scipy.stats import norm, entropy, multivariate_normal, gamma
from scipy import stats 
import torch
from torch.nn import functional as F
from torch import nn, optim
from torch.autograd import Variable
from torch.utils.data import DataLoader
from torch.distributions.gamma import Gamma
from torch.distributions.bernoulli import Bernoulli
from torch.distributions.normal import Normal
from torch.distributions.multivariate_normal import MultivariateNormal
from scipy.stats import entropy 
from scipy import stats
from scipy.sparse import csr_matrix
from sklearn.decomposition import TruncatedSVD
import threading
import subprocess
from scipy.stats import multivariate_normal
import gc
def getClusterPaths(edges, total_stages):
    '''
    Obtain the paths of each cluster for multiple stages.
    
    Args:
    edges: A list of lists, where each sublist contains edges between consecutive stages.
    total_stages: Total number of stages.

    Returns:
    paths: A collection of paths of clusters.
    '''
    if len(edges) != total_stages - 1:
        raise ValueError("Number of edges must be one less than total stages")

    paths = {}

    # Initialize paths with the first set of edges
    for each in edges[0]:
        if str(each[0]) not in paths:
            paths[str(each[0])] = [[each[0]], [each[1]]]
        else:
            paths[str(each[0])][1].append(each[1])

    # Iterate through remaining stages
    for stage in range(1, total_stages - 1):
        for each in edges[stage]:
            for item in paths.keys():
                if len(paths[item]) == stage:
                    continue
                if each[0] in paths[item][stage]:
                    if len(paths[item]) == stage + 1:
                        paths[item].append([each[1]])
                    else:
                        paths[item][stage + 1].append(each[1])

    return paths

def getClusterIdrem(paths, state, total_stages):
    '''
    Concatenate the average gene expression in a cluster tree. Shape: [number of stages, number of genes]
    
    Args: 
    paths: The collection of paths.
    state: A list of average gene expression of each state.
    total_stages: Total number of stages.
    
    Returns: 
    out: A list of gene expression of each cluster tree.
    '''
    out = []

    for path_key in paths.keys():
        path = paths[path_key]

        # Ensure the path contains the expected number of stages
        if len(path) == total_stages:
            stages = [averageNode(node, state[i]) for i, node in enumerate(path)]

            # Reshape each stage and concatenate
            reshaped_stages = [stage.reshape(-1, 1) for stage in stages]
            joint_matrix = np.concatenate(reshaped_stages, axis=1)

            out.append(joint_matrix)

    return out

def averageNode(nodes, state):
    # Dummy implementation of averageNode function. Replace with actual logic.
    # Assuming 'state' is a list of numpy arrays representing gene expression.
    return np.mean([state[node] for node in nodes], axis=0)

def getIdrem(paths,state):
    '''
    concatenate the average gene expression of clusters in each path. shape: [number of stages, number of gene]
    args: 
    paths: the collection of paths
    state: a list of average gene expression of each state
    
    return: 
    out: a list of gene expression of each path
    '''
    out = []
    for i,each in enumerate(paths):
        joint_matrix = np.concatenate((state[0][each[0]].reshape(-1,1), state[1][each[1]].reshape(-1,1), state[2][each[2]].reshape(-1,1), state[3][each[3]].reshape(-1,1)),axis =1)
        out.append(joint_matrix)
    return out

class IDREMthread(threading.Thread):
    def __init__(self, indir, outdir, each,idrem_dir):
        threading.Thread.__init__(self)
        self.indir = indir
        self.outdir = outdir
        self.each = each
        self.idrem_dir = idrem_dir
    def run(self):
        
        command = 'cd '+ str(self.idrem_dir)+' && java -Xmx8G -jar idrem.jar -b %s %s'%(self.indir, self.outdir)
        p = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True)
        print(p.stdout.read())

def runIdrem(paths, midpath, idremInput,genenames,iteration, idrem_dir, species='Human', Minimum_Standard_Deviation = 0.01,Convergence_Likelihood=0.1,Minimum_Absolute_Log_Ratio_Expression=0.05, trained=False):
    '''
    train IDREM model and save the results in iterative training with midpath and iteration
    args:
    paths: the path of IPF progression
    idremInput: average gene expression of each path
    trained: if the model is trained, use saved model
    
    
    
    '''
    dir1 = os.path.join(midpath, str(iteration)+'/idremInput')
    dir2 = os.path.join(midpath, str(iteration)+'/idremsetting')
    dir3 = os.path.join(midpath, str(iteration)+'/idremModel')

    initalcommand = 'mkdir '+dir1+' && mkdir '+dir2+' && mkdir '+dir3
    p = subprocess.Popen(initalcommand, stdout=subprocess.PIPE, shell=True)
    print(p.stdout.read()) 
    print(paths)
    for i, each in enumerate(paths):
        each_processed = []
        for e in each:
            e = str(e).strip('[]')
            e = e.replace(', ','n')
            each_processed.append(e)
        # each_processed = [str(e).strip('[]').replace(', ', 'n') for e in each]
        print(each_processed)
        file_name = '-'.join(each_processed)
        file_path = os.path.join(midpath, str(iteration), 'idremInput', f'{file_name}.txt')
        header = ['gene'] + [f'stage{j}' for j in range(len(each))]
        with open(file_path, 'w') as f:
            f.write('\t'.join(header) + '\n')
            for j, row in enumerate(idremInput[i]):
                row_data = '\t'.join(str(r) for r in row)
                f.write("%s\t%s\n" % (genenames[j], row_data))
        examplefile = open(os.path.join(idrem_dir, 'example_settings.txt'), 'r')
        settings_file_path = os.path.join(midpath, str(iteration), 'idremsetting', f'{file_name}.txt')
        with open(settings_file_path, 'w') as f:
            for k, line in enumerate(examplefile.readlines()):

                if trained and k == 4:
                    print(midpath) 
                    f.write('%s\t%s\n' % ('Saved_Model_File', os.path.join(os.path.abspath(os.path.join(midpath, str(iteration), 'idremInput')), f'{file_name}.txt')))
                elif k == 1:
                    if species == 'Human':
                        f.write('%s\t%s\n' % ('TF-gene_Interaction_Source', 'human_encode.txt.gz'))
                        continue
                elif k == 2:
                    if species == 'Human':
                        f.write('%s\t%s\n' % ('TF-gene_Interactions_File', 'TFInput/human_encode.txt.gz'))
                        continue
                elif k == 5:
                    if species == 'Human':
                        f.write('%s\t%s\n' % ('Gene_Annotation_Source', 'Human (EBI)'))
                        continue
                elif k == 6:
                    if species == 'Human':
                        f.write('%s\t%s\n' % ('Gene_Annotation_File', 'goa_human.gaf.gz'))
                        continue 
                elif k== 17:
                    f.write('%s\n' % ('miRNA-gene_Interaction_Source'))
                    continue
                elif k== 18:
                    f.write('%s\n' % ('miRNA_Expression_Data_File'))
                    continue
                elif k== 26:
                    f.write('%s\n' % ('Proteomics_File'))
                    continue
                elif k == 34:
                    f.write('%s\n' % ('Epigenomic_File'))
                    continue
                elif k == 35:
                    f.write('%s\n' % ('GTF File'))
                    continue
                elif k == 42:
                    f.write('%s\t%s\n' % ('Minimum_Absolute_Log_Ratio_Expression', str(Minimum_Absolute_Log_Ratio_Expression)))
                    continue
                elif k ==51 :
                    f.write('%s\t%s\n' % ('Convergence_Likelihood_%', str(Convergence_Likelihood)))
                    continue
                elif k == 52:
                    f.write('%s\t%s\n' % ('Minimum_Standard_Deviation', str(Minimum_Standard_Deviation)))
                    continue
                elif k != 3:
                    f.write(line)
                else:
                    f.write('%s\t%s\n' % ('Expression_Data_File', os.path.join(os.path.abspath(os.path.join(midpath, str(iteration), 'idremInput')), f'{file_name}.txt')))

        examplefile.close()
        # each0 = str(each[0]).strip('[]')
        # each1 = str(each[1]).strip('[]')
        # each2 = str(each[2]).strip('[]')
        # each3 = str(each[3]).strip('[]')
        
        # each0 = each0.replace(', ','n')
        # each1 = each1.replace(', ','n')
        # each2 = each2.replace(', ','n')
        # each3 = each3.replace(', ','n')
        
        # with open(os.path.join(midpath,str(iteration)+'/idremInput/%s-%s-%s-%s.txt'%(each0,each1,each2,each3)), 'w') as f:
        #     f.write("%s\t%s\t%s\t%s\t%s\n"%('gene','control','IPF1','IPF2','IPF3'))
        #     for j,row in enumerate(idremInput[i]):
        #         f.write("%s\t%s\t%s\t%s\t%s\n" % (genenames[j],str(row[0]),str(row[1]), str(row[2]),str(row[3])))   
        # examplefile = open('./idrem-master/example_settings.txt','r')
    
        # with open(os.path.join(midpath,str(iteration)+'/idremsetting/%s-%s-%s-%s.txt'%(each0,each1,each2,each3)), 'w') as f:
        #     for k,line in enumerate(examplefile.readlines()):
        #         if trained == True:
        #             if k == 4:
        #                 print(os.path.join(midpath,str(iteration)+'/idremInput/%s-%s-%s-%s.txt'%(each0,each1,each2,each3)))
        #                 f.write('%s\t%s\n'%('Saved_Model_File',os.path.join('../'+midpath,str(iteration)+'/idremInput/%s-%s-%s-%s.txt'%(each0,each1,each2,each3))))
        #         if k!= 3:
        #             f.write(line)
        #         else:
        #             f.write('%s\t%s\n'%('Expression_Data_File', os.path.join('../'+midpath,str(iteration)+'/idremInput/%s-%s-%s-%s.txt'%(each0,each1,each2,each3))))
    settinglist = os.listdir(os.path.join(midpath,str(iteration)+'/idremsetting/'))
    
    print(settinglist)
    threads = []
    for each in settinglist:
        if each[0] != '.':
            
            indir = os.path.abspath(os.path.join(midpath,str(iteration)+'/idremsetting/',each))
            outdir =os.path.join(os.path.abspath(os.path.join(midpath,str(iteration))+'/idremModel/'),each)
            
            threads.append(IDREMthread(indir, outdir, each,idrem_dir))
    count = 0
    while True:
        if count<len(threads) and count +2 > len(threads):
            threads[count].start()
            threads[count].join()
            break
        elif count == len(threads):
            break
        else:
            threads[count].start()
            threads[count+1].start()
            threads[count].join()
            threads[count+1].join()
            count+=2
    if not trained:
        print(os.getcwd())
        dir1 = os.path.join(midpath, str(iteration)+'/idremResults')
        dir2 = os.path.join(midpath, str(iteration)+'/idremInput/*.txt_viz')
        command = [['rm -r '+dir1],[ ' mkdir '+dir1], [' mv '+dir2+ ' '+dir1]]
        for each in command:
            p = subprocess.Popen(each, stdout=subprocess.PIPE, shell=True)
            print(p.stdout.read())
    print('idrem Done')

def averageNode(nodes,state):
    '''
    calculate the average gene expression of sibling nodes
    args: 
    nodes: number of sibling nodes
    state: the gene expression of each cluster in a certain stage
    
    return: 
    out: the average gene expression of sibling nodes
    '''
    out = 0
    for each in nodes:
        out+=state[each]
    return out/len(nodes)
# def runIDREMSettings(midpath,iteration):
#     settinglist = os.listdir('./'+midpath+'/'+str(iteration)+'/idremsetting/')
    
#     print(settinglist)
#     threads = []
#     for each in settinglist:
#         if each[0] != '.':
#             indir = os.path.join('../'+midpath+'/'+str(iteration)+'/idremsetting/',each)
#             outdir = os.path.join('../'+midpath+'/'+str(iteration)+'/idremModel/',each)
#             threads.append(IDREMthread(indir, outdir, each))
#     count = 0
#     while True:
#         if count<len(threads) and count +2 > len(threads):
#             threads[count].start()
#             threads[count].join()
#             break
#         elif count == len(threads):
#             break
#         else:
#             threads[count].start()
#             threads[count+1].start()
#             threads[count].join()
#             threads[count+1].join()
#             count+=2

#     command = [['rm -r ./'+midpath+'/'+str(iteration)+'/idremResults'],[ 'mkdir ./'+midpath+'/'+str(iteration)+'/idremResults/'], ['mv ./'+midpath+'/'+str(iteration)+'/idremInput/*.txt_viz ./'+midpath+'/'+str(iteration)+'/idremResults/']]
#     for each in command:
#         p = subprocess.Popen(each, stdout=subprocess.PIPE, shell=True)
#         print(p.stdout.read())
#     print('idrem Done')

if __name__ == '__main__':
    import numpy as np
    edges = [[[2, 0], [0, 1], [8, 3], [3, 4], [3, 5], [4, 6], [5, 7], [6, 8], [1, 9]],  [[1, 0], [4, 1], [3, 2], [0, 3], [7, 5], [0, 6], [8, 7], [10, 10], [3, 11], [8, 12]],[[1, 0], [0, 1], [1, 2], [3, 3], [5, 4], [0, 5], [2, 6], [7, 8], [5, 9], [12, 10], [2, 12]]]
    paths = getClusterPaths(edges,4)
    averageValues = np.load('/mnt/md0/yumin/to_upload/UNAGI/UNAGI/data/example/2/averageValues.npy',allow_pickle=True)
    # print(averageValues.shape)
    idrem= getClusterIdrem(paths,averageValues,4)
    print(idrem)


from UNAGI import UNAGI
unagi = UNAGI()
unagi.analyse_UNAGI('/mnt/md0/yumin/to_upload/UNAGI/data/small/2/stagedata/org_dataset.h5ad',2,10,target_dir=None,customized_drug='/mnt/md0/yumin/to_upload/UNAGI/data/jasper_target_pair.npy')


import numpy as np
import scanpy as sc
import scipy
import os
import pandas as pd
import time
from sklearn.neighbors import kneighbors_graph 
from scipy.stats import norm
from .utils import get_data_file_path
def calculateDataPathwayOverlapGene(adata):
    data_path = get_data_file_path('gesa_pathways.npy')
    pathways = dict(np.load(data_path,allow_pickle=True).tolist())
    # read_path = os.path.join(source_folder,str(ITERATION),'stagedata/dataset.h5ad')
    adata = adata
    genenames = adata.var.index.tolist()
    out = {}
    for each in list(pathways.keys()):
        temp = []
        for gene in pathways[each]:
            if gene in genenames and gene not in temp:
                temp.append(gene)
        if len(temp) >0:
            out[each] = temp
    tmp = {}
    for key, value in out.items():
        value = '!'.join(value)
        if value in tmp:
            tmp[value].append(key)
        else:
            tmp[value] = [ key ]
    out = {}
    for value, keys in tmp.items():
        out[','.join(keys)] = value.split('!')
    # out_path = os.path.join(source_folder,str(ITERATION),'data_pathway_overlap_genes.npy')
    # np.save(out_path,out)
    specific_gene_len_dict = {}
    data_pathway_overlap_genes = out
    for each in list(data_pathway_overlap_genes.keys()):
        specific_gene_len_dict[each] = len(data_pathway_overlap_genes[each])
    adata.uns['data_pathway_overlap_genes'] = out
    adata.uns['pathway_gene_len_dict'] = specific_gene_len_dict
    # adata.write('./iterativeTrainingNOV26/'+str(ITERATION)+'/stagedata/dataset.h5ad')
    # out_path = os.path.join(source_folder,str(ITERATION),'pathway_gene_len_dict.npy')
    # np.save(out_path,specific_gene_len_dict)
    return adata
def calculateTopPathwayGeneRanking(adata):
    '''
    rank pathways based on gene weights
    '''
    pathway_gene = adata.uns['data_pathway_overlap_genes']
    # pathway_gene = np.load('./iterativeTrainingNOV26/'+str(ITERATION)+'/data_pathway_overlap_genes.npy',allow_pickle=True)
    # pathway_gene = dict(pathway_gene.tolist())
    avg_ranking = {}
    for i in range(1,4):
        # adata = sc.read_h5ad(f'./iterativeTrainingNOV26/'+str(ITERATION)+'/stagedata/%d.h5ad'%i)
        stageadataids = adata.obs[adata.obs['stage']==i].index.tolist()
        stageadata = adata[stageadataids]
        avg_ranking[i] = {}
        adata_genes = stageadata.var.index.values.tolist()
        adata_genes_index_dict={}
        for idx,gene in enumerate(adata_genes):
            adata_genes_index_dict[gene] = idx
        for clusterid in set(stageadata.obs['leiden']):
            
            avg_ranking[i][clusterid] = {}
            clusteradataid = stageadata.obs[stageadata.obs['leiden']==clusterid].index.tolist()
            clusteradata = stageadata[clusteradataid]
            cluster_gene_weight_table = clusteradata.layers['geneWeight']
            avg_cluster_gene_weight_table = np.mean(cluster_gene_weight_table,axis=0).reshape(-1)
    
            
            avg_geneWeightTable_ranking = scipy.stats.rankdata(avg_cluster_gene_weight_table)
            avg_geneWeightTable_ranking = np.zeros_like(avg_geneWeightTable_ranking)+len(adata_genes)-avg_geneWeightTable_ranking
    
            for pathway in list(pathway_gene.keys()):
                avg_ranking[i][clusterid][pathway] = 0
                for gene_in_pathway in pathway_gene[pathway]:
                    avg_ranking[i][clusterid][pathway]+=avg_geneWeightTable_ranking[adata_genes_index_dict[gene_in_pathway]]
                avg_ranking[i][clusterid][pathway] = avg_ranking[i][clusterid][pathway]/len(pathway_gene[pathway])
    new_av_ranking = {}
    for i in list(avg_ranking.keys()):
        new_av_ranking[str(i)] = {}
        for j in list(avg_ranking[i].keys()):
            
            new_av_ranking[str(i)][str(j)]={k: idx+1 for idx,(k, v) in enumerate(sorted(avg_ranking[i][j].items(), key=lambda item: item[1]))}
    # np.save('./iterativeTrainingNOV26/'+str(ITERATION)+'/avg_ranking.npy',new_av_ranking)
    adata.uns['pathway_ranking'] = new_av_ranking
    return adata
def findTopGenesInCluster(adata,pval=0.1):
    '''
    find top genes with high gene weight in each cluster (pval < 0.1 default)
    '''
    topClusterGeneBasedOnGeneWeight = {}
    for i in range(1,4):
        stageadataids = adata.obs[adata.obs['stage']==i].index.tolist()
        stageadata = adata[stageadataids]
        topClusterGeneBasedOnGeneWeight[i] = {}
        adata_genes = adata.var.index.values.tolist()
        adata_genes_index_dict={}
        for idx,gene in enumerate(adata_genes):
            adata_genes_index_dict[gene] = idx
        for clusterid in set(stageadata.obs['leiden']):
            topClusterGeneBasedOnGeneWeight[i][clusterid] = []
            clusteradataid = stageadata.obs[stageadata.obs['leiden']==clusterid].index.tolist()
            clusteradata = stageadata[clusteradataid]
            cluster_gene_weight_table = clusteradata.layers['geneWeight']
            avg_cluster_gene_weight_table = np.mean(cluster_gene_weight_table,axis=0).reshape(-1)
            dist_mean = np.mean(avg_cluster_gene_weight_table)
            dist_std = np.std(avg_cluster_gene_weight_table)
            for idx, eachgene in enumerate(avg_cluster_gene_weight_table):
                cdf = 1 - norm.cdf(eachgene, dist_mean, dist_std)
                if cdf < pval:
                    topClusterGeneBasedOnGeneWeight[i][clusterid].append(adata_genes[idx])
    adata.uns['topClusterGeneBasedOnGeneWeight'] = topClusterGeneBasedOnGeneWeight
    return adata
def merge_drugs_with_sametarget_samedirection(adata,overlapped_drug_direction_profile):
    reverse = {}
    for key in list(overlapped_drug_direction_profile.keys()):
        temp = ''
        for value in overlapped_drug_direction_profile[key]:
            temp+=str(value)+','
        temp = temp[:-1]
        if temp not in reverse:
            reverse[temp] = ''+str(key)
        else:
            reverse[temp]+=','+str(key)
    for key in list(reverse.keys()):
        if key == '':
            del reverse[key]
    out = {}
    drug_len = {}
    for key in list(reverse.keys()):
        out[reverse[key]] = []
        for each in key.split(','):
            out[reverse[key]].append(each)
        drug_len[reverse[key]] = len(reverse[key].split(','))
   
    adata.uns['data_drug_overlap_genes'] = out
    adata.uns['drug-gene_len_dict'] = drug_len
    return adata

import gc
from scipy.sparse import csr_matrix
def alterative(source_directory):
    for i in range(4):
        temp1 = sc.read_h5ad(source_directory+'/stagedata/%d.h5ad'%i)
        temp2 = sc.read_h5ad(source_directory+'/stagedata/concat_%d.h5ad'%i)
        temp1.obsp['gcn_connectivities'] = temp2.obsp['gcn_connectivities'].copy()
        temp1.layers['concat'] = csr_matrix(temp2.layers['concat'].copy())
        
        temp2 = None
        gc.collect()
        temp1.write(source_directory+'/stagedata/%d.h5ad'%i, compression='gzip', compression_opts=9)
        

def calculateDrugOverlapGene(adata, cmap_df=None, drug_profile_directory=None):
    genenames = adata.var.index.tolist()
    # use customized drugs
    if drug_profile_directory is not None:
        drug = dict(np.load(drug_profile_directory,
                    allow_pickle=True).tolist())
        out = {}
        for each in list(drug.keys()):
            temp = []
            for gene in drug[each]:
                if gene in genenames and gene not in temp:
                    temp.append(gene)
            if len(temp) > 0:
                out[each] = temp
    
        tmp = {}
        for key, value in out.items():
            value = '!'.join(value)
            if value in tmp:
                tmp[value].append(key)
            else:
                tmp[value] = [key]
        out = {}
        for value, keys in tmp.items():
            out[','.join(keys)] = value.split('!')

        # np.save('./iterativeTrainingNOV26/'+str(ITERATION)+'/data_drug_overlap_genes.npy',out)
        specific_gene_len_dict = {}
        data_pathway_overlap_genes = out
        for each in list(data_pathway_overlap_genes.keys()):
            specific_gene_len_dict[each] = len(
                data_pathway_overlap_genes[each])
  
        adata.uns['data_drug_overlap_genes'] = out
        adata.uns['drug-gene_len_dict'] = specific_gene_len_dict
        return adata
    elif cmap_df is not None:
        cmap_df = cmap_df[cmap_df.columns.intersection(genenames)]
        cols = cmap_df.columns.tolist()
        rows = cmap_df.index.tolist()

        data_pathway_overlap_genes = {row: cols for row in rows}
        specific_gene_len_dict = {compound: len(
            data_pathway_overlap_genes[compound]) for compound in data_pathway_overlap_genes}

        adata.uns['data_drug_overlap_genes'] = data_pathway_overlap_genes
        adata.uns['drug-gene_len_dict'] = specific_gene_len_dict
        # adata.write('./iterativeTrainingNOV26/'+str(ITERATION)+'/stagedata/dataset.h5ad')
        # np.save('./iterativeTrainingNOV26/'+str(ITERATION)+'/drug-gene_len_dict.npy',specific_gene_len_dict)
        return adata


def merge_drugs_with_sametarget_samedirection(adata, overlapped_drug_direction_profile=None):
    reverse = {}
    for key in list(overlapped_drug_direction_profile.keys()):
        temp = ''
        for value in overlapped_drug_direction_profile[key]:
            temp += str(value)+','
        temp = temp[:-1]
        if temp not in reverse:
            reverse[temp] = ''+str(key)
        else:
            reverse[temp] += ','+str(key)
    for key in list(reverse.keys()):
        if key == '':
            del reverse[key]
    out = {}
    drug_len = {}
    for key in list(reverse.keys()):
        out[reverse[key]] = []
        for each in key.split(','):
            out[reverse[key]].append(each)
        drug_len[reverse[key]] = len(reverse[key].split(','))

    adata.uns['data_drug_overlap_genes'] = out
    adata.uns['drug-gene_len_dict'] = drug_len
    return adata


def assign_drug_direction(adata, cmap_df=None, customized_drug_direction=None):
    data_path = get_data_file_path('brdID2cmapName.npy')
    brdID2cmapName = dict(
        np.load(data_path, allow_pickle=True).tolist())
    target_file = adata.uns['data_drug_overlap_genes']
    # assign direciton of drugs to regulate genes based on CMAP direction profile
    if customized_drug_direction is not None:
        CMAP_direction_profile = dict(
            np.load(customized_drug_direction, allow_pickle=True).tolist())
        out = {}
        for each in list(target_file.keys()):  # each='compound1;compound2;'
            lst = each.split(',')  # lst = ['compound1','compound2']
            for item in lst:  # item = 'compound1'
                if item in list(CMAP_direction_profile.keys()):
                    selected_drug = item
                elif brdID2cmapName[item] in list(CMAP_direction_profile.keys()):
                    selected_drug = brdID2cmapName[item]
                else:
                    continue
                temp = CMAP_direction_profile[selected_drug]
                if item not in out:
                    out[item] = []
                for gene in temp:
                    if gene.split(':')[0] in target_file[each]:
                        out[item].append(gene)
        adata = merge_drugs_with_sametarget_samedirection(adata, out)
        return adata
    else:
        out = {}
        
        for each in list(target_file.keys()):  # each='compound1;compound2;'
            lst = each.split(',')  # lst = ['compound1','compound2']
            for item in lst:  # item = 'compound1'
                if item in cmap_df.index.tolist():
                    selected_drug = item
                    break
                elif brdID2cmapName[item] in cmap_df.index.tolist():
                    selected_drug = brdID2cmapName[item]
                    break
                else:
                    continue
            temp = cmap_df.loc[selected_drug]
            temp_record = []
            
            for gene in target_file[each]:
                if gene in cmap_df.columns.tolist():
                    temp_record.append(str(gene)+':'+str(temp[gene]))
            
            out[each] = temp_record
        adata = merge_drugs_with_sametarget_samedirection(adata, out)
        return adata


def find_overlap_and_assign_direction(adata, customized_direction=None, customized_drug=None,cmap_dir=None):
    if customized_drug is None or customized_direction is None:

        arr = np.load(cmap_dir, allow_pickle=True)
        cmap_df = pd.DataFrame(arr[0], index=arr[1], columns=arr[2])

    if customized_drug is None and customized_direction is None:
        print("using cmap drug profile & direction")
        genenames = adata.var.index.tolist()

        cmap_df = cmap_df[cmap_df.columns.intersection(genenames)]
        cols = cmap_df.columns

        out = {row: cols+':'+cmap_df.loc[row]
               for row in cmap_df.index.tolist()}

        nGenes = len(cmap_df.columns)
        drug_len = {row: nGenes for row in cmap_df.index.tolist()}
        adata.uns['drug-gene_len_dict'] = drug_len

        adata = merge_drugs_with_sametarget_samedirection(adata, out)
    else:
        if (customized_drug != None):
            
            print("using customized drug profile:", customized_drug)
            adata = calculateDrugOverlapGene(
                adata, drug_profile_directory=customized_drug)
        else:
            print("using cmap drug profile")
            genenames = adata.var.index.tolist()
            cmap_df = cmap_df[cmap_df.columns.intersection(genenames)]
            adata = calculateDrugOverlapGene(adata, cmap_df=cmap_df)

        if (customized_direction != None):
            print("using cutomized drug profile:", customized_direction)
            adata = assign_drug_direction(
                adata, customized_drug_direction=customized_direction)
        else:
            print("using cmap drug direction")
            adata = assign_drug_direction(adata, cmap_df=cmap_df)

    return adata


import scanpy as sc
import os
import numpy as np
import torch
# from pyro.contrib.examples.util import MNIST
import torch.nn as nn
import torchvision.transforms as transforms
import scanpy as sc
import torch.nn.functional as F
import pyro
import pyro.distributions as dist
# import pyro.contrib.examples.util  # patches torchvision
from pyro.infer import SVI, Trace_ELBO
from pyro.optim import Adam
import matplotlib.pyplot as plt
from torchvision.utils import make_grid
from torch.utils.data import Dataset,DataLoader
from scipy.sparse import csc_matrix
import umap
from .processTFs import mySigmoid
from scvi import data
import scvi
#NUMGENES=len(adata.var)
# customized h5ad dataset
class H5ADataSetTrainClassifier(Dataset):
    def __init__(self,fname):
        self.data=fname
        self.stagelen = len([1,2,3,4])
    def __len__(self):
        return self.data.X.shape[0]
    
    def __getitem__(self,idx):
        x=csc_matrix(self.data.X[idx])[0].toarray()[0]
        x=x.astype(np.float32)
        x_tensor=torch.from_numpy(x)
        label = self.data.obs['stage'].tolist()[idx]
        label = F.one_hot(torch.tensor(label), num_classes=self.stagelen).float()
        return x_tensor,label
    def num_genes(self):
        return len(self.data.var)
    def returnadata(self):
        return self.data
# customized h5ad dataset
class H5ADataSet(Dataset):
    def __init__(self,fname):
        self.data=fname
        # self.neighbors = neighbors
    def __len__(self):
        return self.data.X.shape[0]
    
    def __getitem__(self,idx):

        x=csc_matrix(self.data.X[idx])[0].toarray()[0]
        x=x.astype(np.float32)
        x_tensor=torch.from_numpy(x)
        return x_tensor,self.data.obsp['gcn_connectivities'][idx].indices.tolist(),idx#,self.data.obs['Sample.ID'][idx]
    def num_genes(self):
        return len(self.data.var)
    def returnadata(self):
        return self.data
class plainH5ADataSet(Dataset):
    def __init__(self,fname):
        self.data=fname
        # self.neighbors = neighbors
    def __len__(self):
        return self.data.X.shape[0]
    
    def __getitem__(self,idx):

        x=csc_matrix(self.data.X[idx])[0].toarray()[0]
        x=x.astype(np.float32)
        x_tensor=torch.from_numpy(x)
        return x_tensor
    def num_genes(self):
        return len(self.data.var)
    def returnadata(self):
        return self.data
class H5ADataSetTesting(Dataset):
    def __init__(self,fname):
        self.data=fname
        #self.data.X = self.data.layers['precompute']
    def __len__(self):
        return self.data.X.shape[0]
    
    def __getitem__(self,idx):
        x=csc_matrix(self.data.layers['precompute'][idx])[0].toarray()[0]
        x=x.astype(np.float32)
        x_tensor=torch.from_numpy(x)
        return x_tensor
    def num_genes(self):
        return len(self.data.var)
    def returnadata(self):
        return self.data
class H5ADataSetGeneWeight(Dataset):
    def __init__(self,fname):
        self.data=fname
    
    def __len__(self):
        return self.data.X.shape[0]
    
    def __getitem__(self,idx):
        x=csc_matrix(self.data.X[idx])[0].toarray()[0]
        x=x.astype(np.float32)
        x_tensor=torch.from_numpy(x)
        gw = self.data.layers['geneWeight'][idx][0].toarray()[0]
        gw = mySigmoid(gw.astype(np.float32))
        gw_tensor = torch.from_numpy(gw)
        return x_tensor, gw_tensor
    def num_genes(self):
        return len(self.data.var)
    def returnadata(self):
        return self.data
class H5ADFile:
    def __init__(self,fname):

        self.d1=fname
        
        
    def preprocess(self):
        datControl = self.d1
        #datControl=adata.raw.to_adata()
        sc.pp.filter_cells(datControl, min_genes=200)
        sc.pp.filter_genes(datControl, min_cells=3) 
        #sc.pp.highly_variable_genes(datControl, n_top_genes=5000)
        #datControl = datControl[:, datControl.var.highly_variable]
        datControl.var['mt'] = datControl.var_names.str.startswith('MT-')
        sc.pp.calculate_qc_metrics(datControl, qc_vars=['mt'], percent_top=None, log1p=False, inplace=True)
        sc.pl.violin(datControl, ['n_genes_by_counts', 'total_counts', 'pct_counts_mt'],
                 jitter=0.4, multi_panel=True)
        datControl = datControl[datControl.obs.pct_counts_mt < 5, :]
        #sc.pp.normalize_total(datControl, target_sum=1e4)
        #sc.pp.log1p(datControl)
        self.d1=datControl 
        
    def cluster(self):
        datControl=self.d1
        sc.tl.pca(datControl, svd_solver='arpack')
        sc.pp.neighbors(datControl, n_neighbors=25, n_pcs=40)
        sc.tl.diffmap(datControl)
        sc.tl.leiden(datControl,resolution=0.7)
        sc.tl.paga(datControl)
        sc.pl.paga(datControl,color='leiden')
        sc.tl.umap(datControl,init_pos='paga')
        sc.pl.umap(datControl,color='leiden',legend_loc="on data")
        self.d1=datControl
        
    def getDE(self):
        datControl=self.d1
        sc.tl.rank_genes_groups(datControl, 'leiden', method='wilcoxon',n_genes=100)